Package: bbk
Title: Client for the Bundesbank API
Version: 0.2.0
Authors@R: 
    person("Maximilian", "Muecke", , "muecke.maximilian@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0000-9432-9795"))
Description: Download data and metadata from the 'Bundesbank SDMX Web
    Service API' found at
    <https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-data>.
License: MIT + file LICENSE
URL: https://m-muecke.github.io/bbk/, https://github.com/m-muecke/bbk
BugReports: https://github.com/m-muecke/bbk/issues
Depends: R (>= 4.1.0)
Imports: httr2, stats, utils, xml2
Suggests: ggplot2, scales, testthat (>= 3.0.0), tibble
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.1
NeedsCompilation: no
Packaged: 2024-04-28 09:29:25 UTC; mmuecke
Author: Maximilian Muecke [aut, cre] (<https://orcid.org/0009-0000-9432-9795>)
Maintainer: Maximilian Muecke <muecke.maximilian@gmail.com>
Repository: CRAN
Date/Publication: 2024-04-30 09:20:09 UTC
Built: R 4.2.3; ; 2024-04-30 11:43:26 UTC; unix

Package: bbk
Title: Client for Central Bank APIs
Version: 0.11.0
Authors@R: 
    person("Maximilian", "Mücke", , "muecke.maximilian@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0000-9432-9795"))
Description: A client for retrieving data and metadata from central bank
    APIs including 'Banco de España' (BdE), 'Banco de Portugal' (BdP),
    'Bank for International Settlements' (BIS), 'Bank of Canada' (BoC),
    'Bank of England' (BoE), 'Bank of Japan' (BoJ), 'Banque de France'
    (BdF), 'Czech National Bank' (CNB), 'Deutsche Bundesbank' (BBk),
    'European Central Bank' (ECB), 'National Bank of Poland' (NBP),
    'Norges Bank' (NoB), 'Oesterreichische Nationalbank' (OeNB), 'Sveriges
    Riksbank' (SRb), and 'Swiss National Bank' (SNB).
License: MIT + file LICENSE
URL: https://m-muecke.github.io/bbk/, https://github.com/m-muecke/bbk
BugReports: https://github.com/m-muecke/bbk/issues
Depends: R (>= 4.1.0)
Imports: checkmate, curl, data.table (>= 1.17.0), httr2 (>= 1.0.0),
        jsonlite, stats, utils, xml2
Suggests: ggplot2, knitr, scales, testthat (>= 3.3.0)
Config/Needs/website: knitr, rmarkdown
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-06-11 10:57:02 UTC; mmuecke
Author: Maximilian Mücke [aut, cre] (ORCID:
    <https://orcid.org/0009-0000-9432-9795>)
Maintainer: Maximilian Mücke <muecke.maximilian@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-11 12:10:08 UTC
Built: R 4.5.2; ; 2026-06-11 12:58:08 UTC; unix

Package: bbk
Title: Client for Central Bank APIs
Version: 0.11.0
Authors@R: 
    person("Maximilian", "Mücke", , "muecke.maximilian@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0000-9432-9795"))
Description: A client for retrieving data and metadata from central bank
    APIs including 'Banco de España' (BdE), 'Banco de Portugal' (BdP),
    'Bank for International Settlements' (BIS), 'Bank of Canada' (BoC),
    'Bank of England' (BoE), 'Bank of Japan' (BoJ), 'Banque de France'
    (BdF), 'Czech National Bank' (CNB), 'Deutsche Bundesbank' (BBk),
    'European Central Bank' (ECB), 'National Bank of Poland' (NBP),
    'Norges Bank' (NoB), 'Oesterreichische Nationalbank' (OeNB), 'Sveriges
    Riksbank' (SRb), and 'Swiss National Bank' (SNB).
License: MIT + file LICENSE
URL: https://m-muecke.github.io/bbk/, https://github.com/m-muecke/bbk
BugReports: https://github.com/m-muecke/bbk/issues
Depends: R (>= 4.1.0)
Imports: checkmate, curl, data.table (>= 1.17.0), httr2 (>= 1.0.0),
        jsonlite, stats, utils, xml2
Suggests: ggplot2, knitr, scales, testthat (>= 3.3.0)
Config/Needs/website: knitr, rmarkdown
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-06-11 10:57:02 UTC; mmuecke
Author: Maximilian Mücke [aut, cre] (ORCID:
    <https://orcid.org/0009-0000-9432-9795>)
Maintainer: Maximilian Mücke <muecke.maximilian@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-11 12:10:08 UTC
Built: R 4.6.0; ; 2026-06-11 15:51:14 UTC; unix

# Generated by roxygen2: do not edit by hand

export(bbk_cache_clear)
export(bbk_cache_dir)
export(bbk_data)
export(bbk_dimension)
export(bbk_metadata)
export(bbk_series)
export(bde_data)
export(bde_latest)
export(bdf_codelist)
export(bdf_data)
export(bdf_dataset)
export(bdf_dimension)
export(bdp_data)
export(bdp_dataset)
export(bdp_dimension)
export(bdp_domain)
export(bdp_series)
export(bis_data)
export(bis_dimension)
export(bis_metadata)
export(boc_catalog)
export(boc_data)
export(boc_fx_rates)
export(boc_metadata)
export(boe_data)
export(boj_data)
export(boj_metadata)
export(cnb_czeonia)
export(cnb_data)
export(cnb_dimension)
export(cnb_fx_other_rates)
export(cnb_fx_rates)
export(cnb_indicators)
export(cnb_pribor)
export(cnb_snapshots)
export(cnb_tree)
export(ecb_data)
export(ecb_dimension)
export(ecb_euro_rates)
export(ecb_fx_rates)
export(ecb_metadata)
export(nbp_fx_rates)
export(nbp_gold)
export(nob_data)
export(nob_dimension)
export(nob_metadata)
export(onb_data)
export(onb_dimension)
export(onb_frequency)
export(onb_hierarchy)
export(onb_metadata)
export(onb_toc)
export(snb_data)
export(snb_dimension)
export(snb_metadata)
export(snb_toc)
export(srb_calendar)
export(srb_cross_rates)
export(srb_data)
export(srb_series)
import(checkmate)
import(data.table)
importFrom(httr2,iterate_with_cursor)
importFrom(httr2,req_body_json)
importFrom(httr2,req_cache)
importFrom(httr2,req_error)
importFrom(httr2,req_headers)
importFrom(httr2,req_perform)
importFrom(httr2,req_perform_iterative)
importFrom(httr2,req_retry)
importFrom(httr2,req_url)
importFrom(httr2,req_url_path_append)
importFrom(httr2,req_url_query)
importFrom(httr2,req_user_agent)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(httr2,resp_body_raw)
importFrom(httr2,resp_body_string)
importFrom(httr2,resp_body_xml)
importFrom(httr2,resp_content_type)
importFrom(httr2,resp_status)
importFrom(httr2,resp_url)
importFrom(httr2,resps_data)
importFrom(stats,na.omit)
importFrom(stats,setNames)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# bbk

<!-- badges: start -->

[![R-CMD-check](https://github.com/m-muecke/bbk/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/m-muecke/bbk/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/bbk)](https://CRAN.R-project.org/package=bbk)
[![R-universe
version](https://m-muecke.r-universe.dev/bbk/badges/version)](https://m-muecke.r-universe.dev/bbk)
<!-- badges: end -->

bbk is minimal R client for the following APIs:

- [Banco de España
  (BdE)](https://www.bde.es/webbe/en/estadisticas/recursos/api-estadisticas-bde.html)
- [Banco de Portugal (BdP)](https://bpstat.bportugal.pt/data/docs)
- [Bank for International Settlements
  (BIS)](https://stats.bis.org/api-doc/v1/)
- [Bank of Canada
  (BoC)](https://www.bankofcanada.ca/valet-api-how-to/#about)
- [Bank of England
  (BoE)](https://www.bankofengland.co.uk/boeapps/database)
- [Bank of Japan (BoJ)](https://www.stat-search.boj.or.jp/index_en.html)
- [Banque de France
  (BdF)](https://webstat.banque-france.fr/en/pages/guide-migration-api/)
- [Czech National Bank (CNB)](https://api.cnb.cz/cnbapi/swagger-ui.html)
- [Deutsche Bundesbank
  (BBk)](https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service)
- [European Central Bank
  (ECB)](https://data.ecb.europa.eu/help/api/overview)
- [National Bank of Poland (NBP)](https://api.nbp.pl/en.html)
- [Norges Bank
  (NoB)](https://www.norges-bank.no/en/topics/Statistics/open-data/)
- [Sveriges Riksbank (SRb)](https://developer.api.riksbank.se/)
- [Swiss National Bank (SNB)](https://data.snb.ch/en)
- [Österreichische Nationalbank
  (OeNB)](https://www.oenb.at/en/Statistics/User-Defined-Tables/webservice.html)

> In the future, it may be extended to other central banks and financial
> institutions. Feel free to open an issue if you have a specific
> request.

## Installation

You can install the released version of bbk from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("bbk")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("m-muecke/bbk")
```

## Usage

bbk functions are prefixed according to the central bank they access
(`bbk_`, `ecb_`, `snb_`, `bdp_`, `bis_`, `boe_`, `boj_`, `bde_`, `bdf_`,
`cnb_`, `nbp_`, `onb_`, `boc_`, `nob_`, `srb_`) and follow the naming
conventions of their respective APIs.

The typical workflow involves:

1.  Finding the time series identifier on the relevant central bank’s
    website
2.  Using the appropriate bbk function to retrieve the data

``` r
library(bbk)

# Deutsche Bundesbank - fetch 10 year daily yield curve
yield_curve = bbk_data(
  flow = "BBSIS",
  key = "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A",
  start_period = "2020-01-01"
)
```

<img src="man/figures/README-plotting-1.png" alt="" width="100%" />

## Related work

- [BIS](https://github.com/stefanangrick/BIS): R package for downloading
  data from the Bank for International Settlements (BIS).
- [BISdata](https://github.com/enricoschumann/BISdata): R package for
  downloading data from the Bank for International Settlements (BIS).
- [boe](https://github.com/charlescoverdale/boe): R package for
  downloading data from the Bank of England (BoE).
- [BOJ](https://github.com/stefanangrick/BOJ): R package for downloading
  data from the Bank of Japan (BoJ).
- [SNBdata](https://github.com/enricoschumann/SNBdata): R package for
  downloading data from the Swiss National Bank (SNB).
- [bundesbank](https://github.com/enricoschumann/bundesbank): R scripts
  for downloading time-series data from the Bundesbank.
- [czechrates](https://github.com/jla-data/czechrates): R package for
  Czech National Bank (CNB) foreign exchange and interest rates.
- [ecb](https://github.com/expersso/ecb): R interface to the European
  Central Bank’s Statistical Data Warehouse (SDW) API.
- [oenb](https://github.com/franzmohr/oenb): R package for accessing the
  OeNB’s API.
- [pdfetch](https://github.com/abielr/pdfetch): R package for
  downloading economic and financial time series from public sources.
- [rnbp](https://github.com/szymanskir/rnbp): R package for accessing
  the National Bank of Poland (NBP) API.
- [readsdmx](https://github.com/mdequeljoe/readsdmx): R package for
  reading SDMX data and metadata.
- [rsdmx](https://github.com/eblondel/rsdmx): R package for reading SDMX
  data and metadata.
- [rwebstat](https://github.com/cran/rwebstat): R package for accessing
  the Banque de France’s Webstat API.
- [tidyBdE](https://github.com/rOpenSpain/tidyBdE): R package for
  accessing the Banco de España’s API.
- [valet](https://github.com/runkelcorey/valet): R client to the Bank of
  Canada Valet API.

#' Fetch Bank for International Settlements (BIS) data
#'
#' Retrieve time series data from the BIS SDMX Web Service.
#'
#' @param flow (`character(1)`)\cr
#'   The dataflow to query. See [bis_metadata()] for available dataflows.
#' @param key (`NULL` | `character()`)\cr
#'   The series keys to query using dot-separated dimension values
#'   (e.g., `"M.CH"`). Use `+` for multiple values in one dimension
#'   (e.g., `"M.CH+US"`). If `NULL`, all data for the flow is returned.
#'   Default `NULL`.
#' @param start_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   Start date of the data. Supported formats:
#'   * YYYY for annual data (e.g., `2019`)
#'   * YYYY-S\[1-2\] for semi-annual data (e.g., `"2019-S1"`)
#'   * YYYY-Q\[1-4\] for quarterly data (e.g., `"2019-Q1"`)
#'   * YYYY-MM for monthly data (e.g., `"2019-01"`)
#'   * YYYY-MM-DD for daily data (e.g., `"2019-01-01"`)
#'
#'   If `NULL`, no start date restriction is applied (data retrieved from the earliest available
#'   date). Default `NULL`.
#' @param end_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   End date of the data, in the same format as start_period. If `NULL`, no end date restriction is
#'   applied (data retrieved up to the most recent available date). Default `NULL`.
#' @param first_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the start of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param last_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the end of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param updated_after (`NULL` | `character(1)` | `Date(1)` | `POSIXct(1)`)\cr
#'   Retrieve only observations updated after the given timestamp (e.g.,
#'   `"2024-06-01T00:00:00"`). Useful for incremental retrieval. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://stats.bis.org/api-doc/v1/>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch Swiss central bank policy rate
#' bis_data("WS_CBPOL", "M.CH", last_n = 5L)
#'
#' # fetch effective exchange rates
#' bis_data("WS_EER", "M.N.B.CH", start_period = "2020-01")
#' }
bis_data <- function(
  flow,
  key = NULL,
  start_period = NULL,
  end_period = NULL,
  first_n = NULL,
  last_n = NULL,
  updated_after = NULL
) {
  assert_string(flow, min.chars = 1L)
  assert_character(key, min.chars = 1L, null.ok = TRUE)
  assert_period(start_period)
  assert_period(end_period)
  first_n <- assert_count(first_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  last_n <- assert_count(last_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  updated_after <- assert_timestampish(updated_after, null.ok = TRUE)

  resource <- sdmx_data_resource(flow, key)
  xml <- bis(
    resource,
    accept = "application/vnd.sdmx.genericdata+xml;version=2.1",
    startPeriod = start_period,
    endPeriod = end_period,
    firstNObservations = first_n,
    lastNObservations = last_n,
    updatedAfter = updated_after
  )
  parse_bis_data(xml)
}

#' Fetch Bank for International Settlements (BIS) metadata
#'
#' Retrieve metadata from the BIS SDMX Web Service.
#'
#' @param type (`character(1)`)\cr
#'   The type of metadata to query.
#'   One of: `"datastructure"`, `"dataflow"`, `"codelist"`, or `"concept"`.
#' @param id (`NULL` | `character(1)`)\cr
#'   The id to query. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested metadata.
#' @source <https://stats.bis.org/api-doc/v1/>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bis_metadata("dataflow")
#' bis_metadata("datastructure", "BIS_CBPOL")
#' bis_metadata("codelist", "CL_FREQ")
#' }
bis_metadata <- function(type, id = NULL) {
  assert_choice(type, c("datastructure", "dataflow", "codelist", "concept"))
  assert_string(id, min.chars = 1L, null.ok = TRUE)

  xpath <- switch(
    type,
    datastructure = "//str:DataStructure",
    dataflow = "//str:Dataflow",
    codelist = "//str:Codelist",
    concept = "//str:ConceptScheme"
  )
  type <- if (type == "concept") "conceptscheme" else type
  resource <- if (is.null(id)) type else paste(type, "BIS", toupper(id), sep = "/")
  xml <- bis(resource)
  entries <- xml2::xml_find_all(xml, xpath)
  sdmx_metadata(entries)
}

#' Fetch Bank for International Settlements (BIS) dimensions
#'
#' Retrieve the dimension structure for a given dataflow from the BIS SDMX Web Service.
#'
#' @param id (`character(1)`)\cr
#'   The id of the data structure definition to query (e.g., `"BIS_CBPOL"`).
#' @returns A [data.table::data.table()] with columns:
#'   \item{id}{The dimension id (e.g., `"FREQ"`, `"REF_AREA"`)}
#'   \item{position}{The position of the dimension in the series key}
#'   \item{codelist}{The id of the associated codelist (e.g., `"CL_FREQ"`)}
#' @source <https://stats.bis.org/api-doc/v1/>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bis_dimension("BIS_CBPOL")
#' }
bis_dimension <- function(id) {
  assert_string(id, min.chars = 1L)
  resource <- paste("datastructure", "BIS", toupper(id), sep = "/")
  xml <- bis(resource)
  sdmx_dimension(xml)
}

parse_bis_data <- function(xml) {
  series <- xml2::xml_find_all(xml, ".//generic:Series")
  res <- map(series, function(x) {
    series_key <- x |>
      xml2::xml_find_first(".//generic:SeriesKey") |>
      xml2::xml_children()
    nms <- series_key |>
      xml2::xml_attr("id") |>
      tolower()
    series_key <- series_key |>
      xml2::xml_attr("value") |>
      setNames(nms) |>
      as.list()

    attrs <- x |>
      xml2::xml_find_first(".//generic:Attributes") |>
      xml2::xml_children()
    nms <- attrs |>
      xml2::xml_attr("id") |>
      tolower()
    attrs <- attrs |>
      xml2::xml_attr("value") |>
      setNames(nms) |>
      as.list()

    data <- c(series_key, attrs)
    data$key <- paste(series_key, collapse = ".")
    data$freq <- sdmx_freq(data$freq)
    data$title <- trimws(data$title)

    entries <- xml2::xml_find_all(x, ".//generic:Obs[generic:ObsValue]")
    data$date <- x |>
      xml2::xml_find_all(".//generic:ObsDimension") |>
      xml2::xml_attr("value") |>
      parse_date(data$freq)

    data$value <- entries |>
      xml2::xml_find_all(".//generic:ObsValue") |>
      xml2::xml_attr("value") |>
      as.numeric()

    as.data.table(data)
  })
  res <- res |>
    rbindlist(fill = TRUE) |>
    setcolorder(col_order, skip_absent = TRUE)
  res[]
}

bis_error_body <- function(resp) {
  resp_body_string(resp, "UTF-8")
}

bis <- function(resource, ..., accept = NULL) {
  sdmx_request(
    "https://stats.bis.org/api/v1",
    resource = resource,
    error_body = bis_error_body,
    accept = accept,
    ...
  )
}

test_that("bis_data input validation works", {
  expect_error(bis_data(1L))
  expect_error(bis_data(TRUE))
  expect_error(bis_data(NULL))
  expect_error(bis_data(NA))
  expect_error(bis_data(c("WS_CBPOL", "WS_EER")))
  # key
  expect_error(bis_data("WS_CBPOL", 1L))
  expect_error(bis_data("WS_CBPOL", TRUE))
  expect_error(bis_data("WS_CBPOL", NA))
  # start_period
  expect_error(bis_data("WS_CBPOL", start_period = TRUE))
  expect_error(bis_data("WS_CBPOL", start_period = c("a", "b")))
  # end_period
  expect_error(bis_data("WS_CBPOL", end_period = TRUE))
  expect_error(bis_data("WS_CBPOL", end_period = c("a", "b")))
  # first_n
  expect_error(bis_data("WS_CBPOL", first_n = "abc"))
  expect_error(bis_data("WS_CBPOL", first_n = TRUE))
  expect_error(bis_data("WS_CBPOL", first_n = -1L))
  expect_error(bis_data("WS_CBPOL", first_n = 0L))
  # last_n
  expect_error(bis_data("WS_CBPOL", last_n = "abc"))
  expect_error(bis_data("WS_CBPOL", last_n = TRUE))
  expect_error(bis_data("WS_CBPOL", last_n = -1L))
  expect_error(bis_data("WS_CBPOL", last_n = 0L))
  # updated_after
  expect_error(bis_data("WS_CBPOL", updated_after = 1L))
  expect_error(bis_data("WS_CBPOL", updated_after = TRUE))
  expect_error(bis_data("WS_CBPOL", updated_after = NA))
})

test_that("bis_data passes updated_after as updatedAfter", {
  captured <- NULL
  httr2::local_mocked_responses(function(req) {
    captured <<- req
    httr2::response(200L, headers = "content-type: application/xml", body = charToRaw("<x/>"))
  })
  local_mocked_bindings(parse_bis_data = function(xml) data.table())
  bis_data("WS_CBPOL", "M.CH", updated_after = as.Date("2024-06-01"))
  expect_match(captured$url, "updatedAfter=2024-06-01T00%3A00%3A00")
})

test_that("parse_bis_data works", {
  body <- xml2::read_xml(test_path("fixtures", "bis-data.xml"))
  actual <- parse_bis_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
  expect_numeric(actual$value)
  expect_true(all(c("date", "key", "value", "freq") %in% names(actual)))
  expect_identical(unique(actual$freq), "monthly")
})

test_that("bis_dimension input validation works", {
  expect_error(bis_dimension(1L))
  expect_error(bis_dimension(TRUE))
  expect_error(bis_dimension(NULL))
  expect_error(bis_dimension(c("a", "b")))
})

test_that("bis_metadata input validation works", {
  expect_error(bis_metadata(1L))
  expect_error(bis_metadata(TRUE))
  expect_error(bis_metadata("data"))
  expect_error(bis_metadata(c("datastructure", "dataflow")))
  expect_error(bis_metadata("dataflow", id = 1L))
  expect_error(bis_metadata("dataflow", id = TRUE))
})

test_that("bis_data returns rows from the live endpoint", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  x <- bis_data("WS_CBPOL", "M.CH", last_n = 5L)
  expect_data_table(x, min.rows = 1L)
  expect_true(all(c("date", "key", "value", "freq") %in% names(x)))
})

test_that("sdmx_metadata works for bis", {
  body <- xml2::read_xml(test_path("fixtures", "bis-metadata.xml"))
  entries <- xml2::xml_find_all(body, "//str:Dataflow")
  actual <- sdmx_metadata(entries)
  expect_data_table(actual, min.rows = 1L)
  expect_true(all(c("id", "name") %in% names(actual)))
  expect_true("WS_CBPOL" %in% actual$id)
})

#' Fetch Deutsche Bundesbank (BBk) data
#'
#' Retrieve time series data from the Bundesbank SDMX Web Service.
#'
#' @param flow (`character(1)`)\cr
#'   The flow to query, 5-8 characters. See [bbk_metadata()] for available dataflows.
#' @param key (`NULL` | `character()`)\cr
#'   The series keys to query.
#' @param start_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   The start date of the data. Supported formats:
#'   * YYYY for annual data (e.g., `2019``)
#'   * YYYY-S\[1-2\] for semi-annual data (e.g., `"2019-S1"`)
#'   * YYYY-Q\[1-4\] for quarterly data (e.g., `"2019-Q1"`)
#'   * YYYY-MM for monthly data (e.g., `"2019-01"`)
#'   * YYYY-W\[01-53\] for weekly data (e.g., `"2019-W01"`)
#'   * YYYY-MM-DD for daily and business data (e.g., `"2019-01-01"`)
#'
#'   If `NULL`, no start date restriction is applied (data retrieved from the earliest available
#'   date). Default `NULL`.
#' @param end_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   The end date of the data, in the same format as start_period. If `NULL`, no end date
#'   restriction is applied (data retrieved up to the most recent available date). Default `NULL`.
#' @param first_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the start of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param last_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the end of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param updated_after (`NULL` | `character(1)` | `Date(1)` | `POSIXct(1)`)\cr
#'   Retrieve only observations updated after the given timestamp (e.g.,
#'   `"2024-06-01T00:00:00"`). Useful for incremental retrieval. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-data>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch all data for a given flow and key
#' data <- bbk_data("BBSIS", "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A")
#' head(data)
#'
#' # fetch data for multiple keys
#' data <- bbk_data("BBEX3", c("M.ISK.EUR", "USD.CA.AC.A01"))
#' head(data)
#'
#' # specified period (start date-end date) for daily data
#' data <- bbk_data(
#'   "BBSIS", "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A",
#'   start_period = "2020-01-01",
#'   end_period = "2020-08-01"
#' )
#' head(data)
#'
#' # or only specify the start date
#' data <- bbk_data(
#'   "BBSIS", "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A",
#'   start_period = "2024-04-01"
#' )
#' head(data)
#' }
bbk_data <- function(
  flow,
  key = NULL,
  start_period = NULL,
  end_period = NULL,
  first_n = NULL,
  last_n = NULL,
  updated_after = NULL
) {
  assert_string(flow, min.chars = 5L, max.chars = 8L)
  assert_character(key, min.chars = 1L, null.ok = TRUE)
  assert_period(start_period)
  assert_period(end_period)
  first_n <- assert_count(first_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  last_n <- assert_count(last_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  updated_after <- assert_timestampish(updated_after, null.ok = TRUE)

  resource <- sdmx_data_resource(flow, key)
  xml <- bbk_make_request(
    resource = resource,
    startPeriod = start_period,
    endPeriod = end_period,
    firstNObservations = first_n,
    lastNObservations = last_n,
    preparedAfter = updated_after
  )
  parse_bbk_data(xml)
}

#' Fetch the Deutsche Bundesbank (BBk) series
#'
#' Retrieve a single series by its key via the Bundesbank SDMX Web Service.
#'
#' @inherit bbk_data source
#' @inheritParams bbk_data
#' @inherit bbk_data return
#' @family data
#' @seealso [bbk_data()] for an endpoint with more options.
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bbk_series("BBEX3.M.DKK.EUR.BB.AC.A01")
#' bbk_series("BBAF3.Q.F41.S121.DE.S1.W0.LE.N._X.B")
#' bbk_series("BBBK11.D.TTA000")
#' }
bbk_series <- function(key) {
  assert_string(key, min.chars = 1L)
  body <- bbk_build_request("data/tsIdList", accept = "application/vnd.bbk.data+csv-zip") |>
    req_body_json(key, auto_unbox = FALSE) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_raw()
  parse_bbk_series(body, key)
}

#' Fetch Deutsche Bundesbank (BBk) metadata
#'
#' Retrieve metadata from the Bundesbank time series database via the SDMX Web Service.
#'
#' @param type (`character(1)`)\cr
#'   The type of metadata to query.
#'   One of: `"datastructure"`, `"dataflow"`, `"codelist"`, or `"concept"`.
#' @param id (`NULL` | `character(1)`)\cr
#'   The id to query. Default `NULL`.
#' @param lang (`character(1)`)\cr
#'   Language to query, either `"en"` or `"de"`. Default `"en"`.
#' @returns A [data.table::data.table()] with the requested metadata.
#' The columns are:
#'   \item{id}{The id of the metadata}
#'   \item{name}{The name of the metadata}
#' @source <https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-metadata>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bbk_metadata("datastructure")
#' bbk_metadata("dataflow", "BBSIS")
#' bbk_metadata("codelist", "CL_BBK_ACIP_ASSET_LIABILITY")
#' bbk_metadata("concept", "CS_BBK_BSPL")
#' }
bbk_metadata <- function(type, id = NULL, lang = "en") {
  assert_choice(type, c("datastructure", "dataflow", "codelist", "concept"))
  args <- switch(
    type,
    datastructure = list("datastructure/BBK", "//structure:DataStructure"),
    dataflow = list("dataflow/BBK", "//structure:Dataflow"),
    codelist = list("codelist/BBK", "//structure:Codelist"),
    concept = list("conceptscheme/BBK", "//structure:ConceptScheme")
  )
  dt <- do.call(fetch_bbk_metadata, c(args, list(id, lang)))
  name <- NULL
  dt[!nzchar(name), name := NA_character_][]
}

#' Fetch Deutsche Bundesbank (BBk) dimensions
#'
#' Retrieve the dimension structure for a given dataflow from the Bundesbank SDMX Web Service.
#'
#' @param id (`character(1)`)\cr
#'   The id of the data structure definition to query (e.g., `"BBK_BBSIS"`).
#' @returns A [data.table::data.table()] with columns:
#'   \item{id}{The dimension id (e.g., `"BBK_STD_FREQ"`, `"BBK_STD_AREA"`)}
#'   \item{position}{The position of the dimension in the series key}
#'   \item{codelist}{The id of the associated codelist}
#' @source <https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-metadata>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bbk_dimension("BBK_ERX")
#' }
bbk_dimension <- function(id) {
  assert_string(id, min.chars = 1L)
  resource <- paste("metadata", "datastructure", "BBK", toupper(id), sep = "/")
  xml <- bbk_make_request(resource)
  sdmx_dimension(xml, ns_prefix = "structure")
}

parse_bbk_series <- function(body, key) {
  td <- tempfile()
  dir.create(td)
  on.exit(unlink(td, recursive = TRUE), add = TRUE)
  tf <- file.path(td, "tempfile.zip")
  writeBin(body, tf)
  utils::unzip(tf, exdir = td)

  files <- list.files(td, full.names = TRUE)
  path <- grepv("\\.csv$", files)[[1L]]

  dt <- fread(file = path, header = FALSE, skip = 10L, select = 1:2)
  setnames(dt, c("date", "value"))
  value <- NULL
  dt[value == ".", value := NA_character_]
  dt <- na.omit(dt)

  metadata <- readLines(path, n = 10L)
  title <- sub("^[\",]+", "", metadata[[2L]])
  title <- sub("[\",]+$", "", title)
  freq <- extract_metadata(metadata, "^Time format code")
  unit <- extract_metadata(metadata, "^Unit \\(in english\\),")
  if (is.na(unit)) {
    unit <- extract_metadata(metadata, "^unit,")
  }
  unit_mult <- extract_metadata(metadata, "^unit multiplier,")
  category <- extract_metadata(metadata, "^category,")
  last_update <- extract_metadata(metadata, "^last update,")
  comment <- extract_metadata(metadata, "^Comment \\(in english\\),")
  comment <- sub("^\"", "", comment)
  src <- extract_metadata(metadata, "^Source \\(in english\\),")

  freq <- sdmx_freq(freq)
  dt[, let(
    date = parse_date(date, freq),
    key = key,
    title = title,
    freq = freq,
    category = category,
    unit = unit,
    unit_mult = unit_mult,
    last_update = last_update,
    comment = comment,
    source = src
  )]
  setcolorder(dt, col_order, skip_absent = TRUE)
  dt[]
}

parse_bbk_data <- function(xml) {
  series <- xml2::xml_find_all(xml, ".//generic:Series")
  dt <- rbindlist(map(series, function(x) {
    series_key <- x |>
      xml2::xml_find_first(".//generic:SeriesKey") |>
      xml2::xml_children()
    nms <- series_key |>
      xml2::xml_attr("id") |>
      tolower()
    series_key <- series_key |>
      xml2::xml_attr("value") |>
      setNames(nms) |>
      as.list()

    attrs <- x |>
      xml2::xml_find_first(".//generic:Attributes") |>
      xml2::xml_children()
    nms <- attrs |>
      xml2::xml_attr("id") |>
      tolower()
    attrs <- attrs |>
      xml2::xml_attr("value") |>
      setNames(nms) |>
      as.list()

    data <- c(series_key, attrs)
    nms <- names(data)
    nms <- sub("^bbk_(seis_)?", "", nms)
    nms <- sub("^std_", "", nms)
    has_eng <- paste0(nms, "_eng") %in% nms
    data <- data[!has_eng]
    nms <- sub("_eng$", "", nms[!has_eng])
    # fmt: skip
    nms <- fcase(
      nms == "id", "key",
      nms == "web_category", "category",
      default = nms
    )
    names(data) <- nms

    data$freq <- sdmx_freq(data$time_format)

    entries <- xml2::xml_find_all(xml, "//generic:Obs[generic:ObsValue]")
    data$date <- entries |>
      xml2::xml_find_all(".//generic:ObsDimension") |>
      xml2::xml_attr("value") |>
      parse_date(data$freq)

    data$value <- entries |>
      xml2::xml_find_all(".//generic:ObsValue") |>
      xml2::xml_attr("value") |>
      as.numeric()

    as.data.table(data)
  }))
  decimals <- NULL
  dt[, decimals := as.integer(decimals)]
  setcolorder(dt, col_order, skip_absent = TRUE)
  dt[]
}

fetch_bbk_metadata <- function(resource, xpath, id = NULL, lang = "en") {
  assert_choice(lang, c("en", "de"))
  assert_string(id, min.chars = 1L, null.ok = TRUE)

  resource <- paste("metadata", resource, sep = "/")
  if (!is.null(id)) {
    resource <- paste(resource, toupper(id), sep = "/")
  }
  xml <- bbk_make_request(resource)
  entries <- xml2::xml_find_all(xml, xpath)
  sdmx_metadata(entries, lang, ns_prefix = "common")
}

bbk_error_body <- function(resp) {
  content_type <- resp_content_type(resp)
  if (identical(content_type, "application/json")) {
    msg <- resp_body_json(resp)$title
    docs <- "See docs at <https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/status-codes/status-codes-855918>" # nolint
    c(msg, docs)
  }
}

bbk_build_request <- function(resource, accept = NULL) {
  request("https://api.statistiken.bundesbank.de/rest") |>
    req_user_agent(bbk_user_agent()) |>
    req_headers(`Accept-Language` = "en", accept = accept) |>
    req_url_path_append(resource) |>
    req_error(body = bbk_error_body)
}

bbk_make_request <- function(resource, ...) {
  sdmx_request("https://api.statistiken.bundesbank.de/rest", resource, bbk_error_body, ...)
}

test_that("bbk_data input validation works", {
  # flow should be a character(1) with 5 <= nchar <= 8
  expect_error(bbk_data(1L, "abc"))
  expect_error(bbk_data(TRUE, "abc"))
  expect_error(bbk_data(NULL, "abc"))
  expect_error(bbk_data(c("a", "b"), "abc"))
  expect_error(bbk_data("a", "abc"))
  expect_error(bbk_data("abcd", "abc"))
  # key should be a character()
  expect_error(bbk_data("abcde", 1L))
  expect_error(bbk_data("abcde", TRUE))
  expect_error(bbk_data("abcde", NULL))
  # start_period should be a character(1) or NULL
  expect_error(bbk_data("abcde", "abc", start_period = 1L))
  expect_error(bbk_data("abcde", "abc", start_period = TRUE))
  expect_error(bbk_data("abcde", "abc", start_period = c("a", "b")))
  # end_period should be a character(1) or NULL
  expect_error(bbk_data("abcde", "abc", TRUE))
  expect_error(bbk_data("abcde", "abc", end_period = 1L))
  expect_error(bbk_data("abcde", "abc", end_period = c("a", "b")))
  # first_n should be a whole number or NULL
  expect_error(bbk_data("abcde", "abc", first_n = "abc"))
  expect_error(bbk_data("abcde", "abc", first_n = 1:2))
  expect_error(bbk_data("abcde", "abc", first_n = TRUE))
  expect_error(bbk_data("abcde", "abc", first_n = 1.5))
  expect_error(bbk_data("abcde", "abc", first_n = -1L))
  expect_error(bbk_data("abcde", "abc", first_n = 0L))
  # last_n should be a whole number or NULL
  expect_error(bbk_data("abcde", "abc", last_n = "abc"))
  expect_error(bbk_data("abcde", "abc", last_n = TRUE))
  expect_error(bbk_data("abcde", "abc", last_n = 1:2))
  expect_error(bbk_data("abcde", "abc", last_n = 1.5))
  expect_error(bbk_data("abcde", "abc", last_n = -1L))
  expect_error(bbk_data("abcde", "abc", last_n = 0L))
  # updated_after should be NULL, Date, POSIXct, or ISO 8601 string
  expect_error(bbk_data("abcde", "abc", updated_after = 1L))
  expect_error(bbk_data("abcde", "abc", updated_after = TRUE))
  expect_error(bbk_data("abcde", "abc", updated_after = NA))
})

test_that("bbk_data passes updated_after as preparedAfter", {
  captured <- NULL
  httr2::local_mocked_responses(function(req) {
    captured <<- req
    httr2::response(200L, headers = "content-type: application/xml", body = charToRaw("<x/>"))
  })
  local_mocked_bindings(parse_bbk_data = function(xml) data.table())
  bbk_data("BBSIS", "abc", updated_after = as.Date("2024-06-01"))
  expect_match(captured$url, "preparedAfter=2024-06-01T00%3A00%3A00")
})

test_that("parse_bbk_data works", {
  body <- xml2::read_xml(test_path("fixtures", "bbk-data.xml"))
  actual <- parse_bbk_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_identical(
    unique(actual$key),
    "BBSIS.D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A"
  )
  expect_date(actual$date)
})

test_that("parse_bbk_series works", {
  body <- readRDS(test_path("fixtures", "bbk-series.rds"))
  actual <- parse_bbk_series(
    body,
    "BBSIS.D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A"
  )
  nms <- c(
    "key",
    "date",
    "value",
    "title",
    "freq",
    "category",
    "unit",
    "unit_mult",
    "last_update",
    "comment",
    "source"
  )
  expect_data_table(actual, min.rows = 1L)
  expect_subset(names(actual), nms)
  expect_identical(
    unique(actual$key),
    "BBSIS.D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A"
  )
  expect_date(actual$date)
})

test_that("bbk_series input validation works", {
  expect_error(bbk_series(1L))
  expect_error(bbk_series(character()))
  expect_error(bbk_series(NA))
  expect_error(bbk_series())
  expect_error(bbk_series(NULL))
  expect_error(bbk_series(c("BBEX3.M.DKK.EUR.BB.AC.A01", "BBEX3.M.DKK.EUR.BB.AC.A02")))
})

test_that("bbk_dimension input validation works", {
  expect_error(bbk_dimension(1L))
  expect_error(bbk_dimension(TRUE))
  expect_error(bbk_dimension(NULL))
  expect_error(bbk_dimension(c("a", "b")))
})

test_that("metadata input validation works", {
  # type should be one of datastructure, dataflow, codelist, concept
  expect_error(bbk_metadata(1L))
  expect_error(bbk_metadata(TRUE))
  expect_error(bbk_metadata(c("datastructure", "dataflow")))
  expect_error(bbk_metadata("data"))
  # id should be a character(1) or NULL
  expect_error(bbk_metadata("datastructure", id = 1L))
  expect_error(bbk_metadata("datastructure", id = TRUE))
  expect_error(bbk_metadata("datastructure", id = c("a", "b")))
  # lang should be en or de
  expect_error(bbk_metadata(lang = "abc"))
  expect_error(bbk_metadata(lang = 1L))
  expect_error(bbk_metadata(lang = NA))
})

test_that("bbk_series does frequency conversion", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  x <- bbk_series("BBEX3.M.DKK.EUR.BB.AC.A01")
  expect_all_equal(x$freq, "monthly")
  x <- bbk_series("BBAF3.Q.F41.S121.DE.S1.W0.LE.N._X.B")
  expect_all_equal(x$freq, "quarterly")
  x <- bbk_series("BBBK11.D.TTA000")
  expect_all_equal(x$freq, "daily")
})

#' Fetch Österreichische Nationalbank (OeNB) data
#'
#' Retrieve time series data from the OeNB Web Service.
#'
#' @param hier_id (`integer(1)`)\cr
#'   Hierarchy id to query.
#' @param key (`character()`)\cr
#'   The series keys to query.
#' @param ... (`any`)\cr
#'   Additional parameters to pass to the API.
#' @param start_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   Start date of the data.
#' @param end_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   End date of the data.
#' @param freq (`NULL` | `character(1)`)\cr
#'   Frequency of the data.
#' @param lang (`character(1)`)\cr
#'   Language to query. Default `"en"`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.oenb.at/en/Statistics/User-Defined-Tables/webservice.html>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' onb_data(hier_id = 11, key = "VDBFKBSC217000")
#'
#' # Loans to euro area residents, since 2000:
#' onb_data(hier_id = 11, key = "VDBFKBSC217000", start_period = "2000-01-01")
#'
#' # Austrian imports and exports of goods from/to Germany, 2002–2012, annual frequency:
#' onb_data(hier_id = 901, key = "VDBQZA1000", start_period = 2002, end_period = 2012, freq = "A")
#'
#' # Number of Austrian banks' subsidiaries abroad and in the EU, from 2005, semiannual:
#' onb_data(
#'   hier_id = 321,
#'   key = c("VDBKISDANZTAU", "VDBKISDANZTEU"),
#'   start_period = 200501,
#'   freq = "H"
#' )
#' }
onb_data <- function(
  hier_id,
  key,
  ...,
  start_period = NULL,
  end_period = NULL,
  freq = NULL,
  lang = "en"
) {
  hier_id <- assert_count(hier_id, positive = TRUE, coerce = TRUE)
  assert_character(key, min.chars = 1L)
  assert_period(start_period)
  assert_period(end_period)
  assert_string(freq, min.chars = 1L, null.ok = TRUE)
  assert_string(lang, n.chars = 2L)

  xml <- onb(
    resource = "data",
    hierid = hier_id,
    pos = key,
    starttime = start_period,
    endtime = end_period,
    freq = freq,
    lang = toupper(lang),
    ...
  )
  parse_onb_data(xml)
}

parse_onb_data <- function(xml) {
  dt <- xml |>
    xml2::xml_find_all(".//dataSet") |>
    map(function(x) {
      obs <- xml2::xml_find_all(x, ".//obs")
      dt <- data.table(
        date = xml2::xml_attr(obs, "periode"),
        value = as.numeric(xml2::xml_attr(obs, "value"))
      )
      attrs <- xml2::xml_attrs(x)
      dt[, names(attrs) := as.list(attrs)]
    }) |>
    rbindlist() |>
    setnames(convert_camel_case) |>
    setnames(c("pos", "pos_title"), c("key", "title")) |>
    setcolorder(col_order, skip_absent = TRUE)
  dt[]
}

#' Fetch Österreichische Nationalbank (OeNB) metadata
#'
#' @inheritParams onb_data
#' @inherit onb_data return
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' onb_metadata(hier_id = 11, key = "VDBFKBSC217000")
#' }
onb_metadata <- function(hier_id, key, ..., lang = "en") {
  hier_id <- assert_count(hier_id, positive = TRUE, coerce = TRUE)
  assert_string(key, min.chars = 1L)
  assert_string(lang, n.chars = 2L)

  xml <- onb(resource = "meta", hierid = hier_id, pos = key, lang = toupper(lang), ...)
  parse_onb_metadata(xml)
}

parse_onb_metadata <- function(xml) {
  meta <- xml |> xml2::xml_find_all("meta") |> xml2::xml_children()
  x <- setNames(xml2::as_list(meta), xml2::xml_name(meta))
  x <- x[lengths(x) == 1L]
  x <- unlist(x, recursive = FALSE)
  dt <- setDT(x)
  dt[, names(dt) := map(.SD, trimws), .SDcols = is.character]
  dt[, names(dt) := map(.SD, \(x) replace(x, x == "-", NA)), .SDcols = is.character][]
}

#' Fetch Österreichische Nationalbank (OeNB) data frequency
#'
#' @inheritParams onb_data
#' @inherit onb_data return
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' onb_frequency(hier_id = 74, key = "VDBOSBHAGBSTIN")
#' onb_frequency(hier_id = 11, key = "VDBFKBSC217000")
#' }
onb_frequency <- function(hier_id, key, ..., lang = "en") {
  hier_id <- assert_count(hier_id, positive = TRUE, coerce = TRUE)
  assert_string(key, min.chars = 1L, null.ok = TRUE)
  assert_string(lang, n.chars = 2L)

  xml <- onb(resource = "datafrequency", hierid = hier_id, pos = key, lang = toupper(lang), ...)
  parse_onb_frequency(xml)
}

parse_onb_frequency <- function(xml) {
  dt <- xml |>
    xml2::xml_find_all(".//dataSet") |>
    map(function(x) {
      freq <- x |> xml2::xml_find_all(".//periods") |> xml2::xml_attr("frequency")
      avail <- x |> xml2::xml_find_all(".//periods/available") |> xml2::xml_text()
      dt <- data.table(freq = freq, available = avail)
      attrs <- xml2::xml_attrs(x)
      dt[, names(attrs) := as.list(attrs)]
    }) |>
    rbindlist() |>
    setnames(\(x) sub("_code", "", tolower(x), fixed = TRUE))
  dt
}

#' Fetch Österreichische Nationalbank (OeNB) table of contents
#'
#' @inheritParams onb_data
#' @inherit onb_data return
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' onb_toc()
#' }
onb_toc <- function(lang = "en") {
  assert_string(lang, n.chars = 2L)
  xml <- onb(resource = "content", lang = toupper(lang))
  parse_onb_toc(xml)
}

#' Fetch Österreichische Nationalbank (OeNB) hierarchy
#'
#' @inheritParams onb_data
#' @inherit onb_data return
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' onb_hierarchy(hier_id = 11)
#' }
onb_hierarchy <- function(hier_id, lang = "en") {
  hier_id <- assert_count(hier_id, positive = TRUE, coerce = TRUE)
  xml <- onb(resource = "content", hierid = hier_id, lang = toupper(lang))
  parse_onb_hierarchy(xml)
}

#' Fetch Österreichische Nationalbank (OeNB) dimension
#'
#' @inheritParams onb_data
#' @inherit onb_data return
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' onb_dimension(hier_id = 11, key = "VDBFKBSC217000")
#' }
onb_dimension <- function(hier_id, key, lang = "en") {
  hier_id <- assert_count(hier_id, positive = TRUE, coerce = TRUE)
  assert_string(key, min.chars = 1L)
  assert_string(lang, n.chars = 2L)

  xml <- onb(resource = "content", hierid = hier_id, pos = key, lang = toupper(lang))
  parse_onb_dimension(xml)
}

parse_onb_toc <- function(xml) {
  elem <- xml2::xml_find_all(xml, ".//content/element")
  dt <- elem |>
    xml2::xml_attrs() |>
    map(\(x) setDT(as.list(x))) |>
    rbindlist()
  desc <- xml2::xml_text(xml2::xml_find_all(elem, "text"))
  id <- parent <- NULL
  dt[, let(id = as.integer(id), parent = as.integer(parent), description = desc)]
  dt[]
}

parse_onb_hierarchy <- function(xml) {
  xml |>
    xml2::xml_find_all(".//group") |>
    map(function(grp) {
      pos <- xml2::xml_find_all(grp, ".//position")
      data.table(
        group = xml2::xml_attr(grp, "name"),
        id = xml2::xml_attr(pos, "id"),
        value = xml2::xml_text(pos)
      )
    }) |>
    rbindlist()
}

parse_onb_dimension <- function(xml) {
  dt <- xml |>
    xml2::xml_find_all(".//data_content/structure/dimension") |>
    xml2::xml_attrs() |>
    map(\(x) setDT(as.list(x))) |>
    rbindlist()
  nr <- NULL
  dt[, nr := as.integer(nr)][]
}

onb <- function(resource, ...) {
  request("https://www.oenb.at/isadataservice") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(resource) |>
    req_url_query(..., .multi = "explode") |>
    req_error(
      is_error = function(resp) {
        if (resp_status(resp) >= 400L) {
          return(TRUE)
        }
        x <- resp |> resp_body_xml() |> xml2::xml_find_first("errors")
        !is.na(x)
      },
      body = onb_error_body
    ) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_xml()
}


onb_error_body <- function(resp) {
  content_type <- resp_content_type(resp)
  if (identical(content_type, "application/xml")) {
    xml2::xml_text(resp_body_xml(resp))
  }
}

#' @keywords internal
#' @import checkmate
#' @import data.table
#' @importFrom httr2 iterate_with_cursor
#' @importFrom httr2 req_body_json
#' @importFrom httr2 req_cache
#' @importFrom httr2 req_error
#' @importFrom httr2 req_headers
#' @importFrom httr2 req_perform
#' @importFrom httr2 req_perform_iterative
#' @importFrom httr2 req_retry
#' @importFrom httr2 req_url
#' @importFrom httr2 req_url_path_append
#' @importFrom httr2 req_url_query
#' @importFrom httr2 req_user_agent
#' @importFrom httr2 request
#' @importFrom httr2 resp_body_json
#' @importFrom httr2 resp_body_raw
#' @importFrom httr2 resp_body_string
#' @importFrom httr2 resp_body_xml
#' @importFrom httr2 resp_content_type
#' @importFrom httr2 resp_status
#' @importFrom httr2 resp_url
#' @importFrom httr2 resps_data
#' @importFrom stats na.omit setNames
"_PACKAGE"

col_order <- c("date", "key", "value", "freq", "title", "description")

bbk_user_agent <- function() {
  sprintf("bbk/%s", utils::packageVersion("bbk"))
}

#' Fetch Banco de Portugal (BdP) data
#'
#' Retrieve time series data from the BPstat API.
#'
#' @details
#' The BPstat API uses a two-step workflow: first look up the series metadata with
#' [bdp_series()] to find the `domain_id` and `dataset_id`, then use those to fetch the actual
#' observations.
#'
#' You can browse available series at the [BPstat portal](https://bpstat.bportugal.pt).
#'
#' @param domain_id (`integer(1)`)\cr
#'   The domain ID. Use [bdp_domain()] to list available domains.
#' @param dataset_id (`character(1)`)\cr
#'   The dataset ID within the domain.
#' @param series_ids (`NULL` | `integer()`)\cr
#'   Optional series IDs to filter the dataset.
#' @param start_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   Start date of the data.
#' @param end_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   End date of the data.
#' @param last_n (`NULL` | `integer(1)`)\cr
#'   Return only the last `n` observations per series.
#' @param updated_after (`NULL` | `character(1)` | `Date(1)` | `POSIXct(1)`)\cr
#'   Retrieve only observations published after the given timestamp (e.g.,
#'   `"2024-06-01T00:00:00"`). Useful for incremental retrieval. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param lang (`character(1)`)\cr
#'   Language for labels, either `"en"` or `"pt"`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://bpstat.bportugal.pt/data/docs>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # Portuguese GDP (annual, current prices)
#' bdp_data(54L, "ce3e4e50cda325537eff729ef64037cd", series_ids = 12518356L)
#' }
bdp_data <- function(
  domain_id,
  dataset_id,
  series_ids = NULL,
  start_date = NULL,
  end_date = NULL,
  last_n = NULL,
  updated_after = NULL,
  lang = "en"
) {
  domain_id <- assert_count(domain_id, positive = TRUE, coerce = TRUE)
  assert_string(dataset_id, min.chars = 1L)
  assert_integerish(series_ids, lower = 1L, null.ok = TRUE)
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)
  last_n <- assert_count(last_n, positive = TRUE, null.ok = TRUE, coerce = TRUE)
  updated_after <- assert_timestampish(updated_after, null.ok = TRUE)
  assert_choice(lang, c("en", "pt"))

  json <- bdp(
    "domains",
    domain_id,
    "datasets",
    dataset_id,
    lang = lang,
    series_ids = series_ids,
    obs_since = start_date,
    obs_to = end_date,
    obs_last_n = last_n,
    obs_published_since = updated_after
  )
  parse_bdp_data(json)
}

parse_bdp_data <- function(json) {
  time_dim <- json$role$time[[1L]]
  dates <- unlist(json$dimension[[time_dim]]$category$index, use.names = FALSE)
  n_dates <- length(dates)
  values <- unlist(json$value, use.names = FALSE)
  n_series <- length(values) %/% n_dates

  dt <- data.table(
    date = as.Date(rep(dates, times = n_series)),
    value = values,
    status = unlist(json$status, use.names = FALSE)
  )

  series <- json$extension$series
  if (!is.null(series)) {
    dt[, "key" := rep(map_int(series, "id"), each = n_dates)]
    dt[, "title" := rep(map_chr(series, "label"), each = n_dates)]
  }

  dt[, "freq" := bdp_freq(dates)]
  setcolorder(dt, col_order, skip_absent = TRUE)
  dt[]
}

bdp_freq <- function(dates) {
  if (length(dates) < 2L) {
    return("annual")
  }
  d <- as.Date(dates)
  diff_days <- as.integer(diff(d[1:2]))
  if (diff_days <= 1L) {
    "daily"
  } else if (diff_days <= 35L) {
    "monthly"
  } else if (diff_days <= 100L) {
    "quarterly"
  } else {
    "annual"
  }
}

#' Fetch Banco de Portugal (BdP) series metadata
#'
#' Retrieve metadata for one or more series from the BPstat API. This is useful to discover the
#' `domain_id` and `dataset_id` needed for [bdp_data()].
#'
#' @param series_ids (`integer()`)\cr
#'   One or more series IDs to look up.
#' @param lang (`character(1)`)\cr
#'   Language for labels, either `"en"` or `"pt"`.
#' @returns A [data.table::data.table()] with series metadata including `domain_id` and
#'   `dataset_id`.
#' @source <https://bpstat.bportugal.pt/data/docs>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bdp_series(12518356L)
#' }
bdp_series <- function(series_ids, lang = "en") {
  assert_integerish(series_ids, lower = 1L, min.len = 1L)
  assert_choice(lang, c("en", "pt"))

  json <- bdp("series", lang = lang, series_ids = series_ids)
  parse_bdp_series(json)
}

parse_bdp_series <- function(json) {
  dt <- data.table(
    id = map_int(json, "id"),
    label = map_chr(json, "label"),
    short_label = map_chr(json, "short_label"),
    description = map_chr(json, "description"),
    dataset_id = map_chr(json, "dataset_id"),
    domain_id = map_int(json, \(x) x$domain_ids[[1L]]),
    obs_updated_at = map_chr(json, "obs_updated_at")
  )
  obs_updated_at <- NULL
  dt[, obs_updated_at := as.POSIXct(obs_updated_at, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")]
  dt[]
}

#' Fetch Banco de Portugal (BdP) datasets
#'
#' Retrieve the list of datasets for a given domain from the BPstat API.
#'
#' @inheritParams bdp_data
#' @returns A [data.table::data.table()] with available datasets.
#' @source <https://bpstat.bportugal.pt/data/docs>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bdp_dataset(54L)
#' }
bdp_dataset <- function(domain_id, lang = "en") {
  domain_id <- assert_count(domain_id, positive = TRUE, coerce = TRUE)
  assert_choice(lang, c("en", "pt"))

  req <- bdp_request("domains", domain_id, "datasets", lang = lang)
  items <- req |>
    req_perform_iterative(next_req = bdp_next_req, max_reqs = Inf) |>
    resps_data(\(resp) resp_body_json(resp)$link$item)
  parse_bdp_dataset(items)
}

bdp_next_req <- function(resp, req) {
  next_url <- resp_body_json(resp)$extension$next_page
  if (is.null(next_url)) {
    return()
  }
  req_url(req, next_url)
}

parse_bdp_dataset <- function(items) {
  data.table(
    id = map_chr(items, \(x) x$extension$id),
    label = map_chr(items, "label"),
    num_series = map_int(items, \(x) x$extension$num_series),
    obs_updated_at = as.POSIXct(
      map_chr(items, \(x) x$extension$obs_updated_at),
      format = "%Y-%m-%dT%H:%M:%SZ",
      tz = "UTC"
    )
  )
}

#' Fetch Banco de Portugal (BdP) dimensions
#'
#' Retrieve the list of dimensions for a given domain, or the categories within a single dimension.
#'
#' @inheritParams bdp_data
#' @param dimension_id (`NULL` | `integer(1)`)\cr
#'   Optional dimension ID. If `NULL`, all dimensions for the domain are returned. If specified,
#'   the categories within that dimension are returned.
#' @returns A [data.table::data.table()] with dimensions or categories.
#' @source <https://bpstat.bportugal.pt/data/docs>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bdp_dimension(54L)
#' }
bdp_dimension <- function(domain_id, dimension_id = NULL, lang = "en") {
  domain_id <- assert_count(domain_id, positive = TRUE, coerce = TRUE)
  dimension_id <- assert_count(dimension_id, positive = TRUE, null.ok = TRUE, coerce = TRUE)
  assert_choice(lang, c("en", "pt"))

  if (is.null(dimension_id)) {
    req <- bdp_request("domains", domain_id, "dimensions", lang = lang)
    items <- req |>
      req_perform_iterative(next_req = bdp_next_req, max_reqs = Inf) |>
      resps_data(\(resp) resp_body_json(resp)$link$item)
    parse_bdp_dimension(items)
  } else {
    json <- bdp("domains", domain_id, "dimensions", dimension_id, lang = lang)
    parse_bdp_category(json)
  }
}

parse_bdp_dimension <- function(items) {
  data.table(
    id = map_int(items, \(x) x$extension$id),
    label = map_chr(items, "label"),
    description = map_chr(items, \(x) x$extension$description)
  )
}

parse_bdp_category <- function(json) {
  labels <- json$category$label
  data.table(
    id = as.integer(names(labels)),
    label = unlist(labels, use.names = FALSE)
  )
}

#' Fetch Banco de Portugal (BdP) domains
#'
#' Retrieve the list of available statistical domains from the BPstat API, or details for a single
#' domain.
#'
#' @param domain_id (`NULL` | `integer(1)`)\cr
#'   Optional domain ID. If `NULL`, all domains are returned.
#' @inheritParams bdp_series
#' @returns A [data.table::data.table()] with available domains.
#' @source <https://bpstat.bportugal.pt/data/docs>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bdp_domain()
#' }
bdp_domain <- function(domain_id = NULL, lang = "en") {
  domain_id <- assert_count(domain_id, positive = TRUE, null.ok = TRUE, coerce = TRUE)
  assert_choice(lang, c("en", "pt"))

  json <- bdp("domains", domain_id, lang = lang)
  if (!is.null(domain_id)) {
    json <- list(json)
  }
  parse_bdp_domain(json)
}

parse_bdp_domain <- function(json) {
  data.table(
    id = map_int(json, "id"),
    parent_id = map_int(json, \(x) x$parent_id %??% NA_integer_),
    label = map_chr(json, "label"),
    short_label = map_chr(json, "short_label"),
    has_series = map_lgl(json, "has_series"),
    num_series = map_int(json, \(x) x$num_series %??% NA_integer_),
    num_datasets = map_int(json, \(x) x$num_datasets %??% NA_integer_)
  )
}

bdp <- function(
  ...,
  lang = "en",
  series_ids = NULL,
  obs_since = NULL,
  obs_to = NULL,
  obs_last_n = NULL,
  obs_published_since = NULL
) {
  bdp_request(
    ...,
    lang = lang,
    series_ids = series_ids,
    obs_since = obs_since,
    obs_to = obs_to,
    obs_last_n = obs_last_n,
    obs_published_since = obs_published_since
  ) |>
    req_perform() |>
    resp_body_json()
}

bdp_request <- function(
  ...,
  lang = "en",
  series_ids = NULL,
  obs_since = NULL,
  obs_to = NULL,
  obs_last_n = NULL,
  obs_published_since = NULL
) {
  request("https://bpstat.bportugal.pt/data/v1") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(...) |>
    req_url_query(
      lang = toupper(lang),
      series_ids = series_ids,
      obs_since = obs_since,
      obs_to = obs_to,
      obs_last_n = obs_last_n,
      obs_published_since = obs_published_since,
      .multi = "comma"
    ) |>
    req_error(body = bdp_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache()
}

bdp_error_body <- function(resp) {
  content_type <- resp_content_type(resp)
  if (identical(content_type, "application/json")) {
    json <- resp_body_json(resp)
    msg <- json$detail %??% json$message
    docs <- "See docs at <https://bpstat.bportugal.pt/data/docs>"
    c(msg, docs)
  }
}

test_that("bdp_data input validation works", {
  expect_error(bdp_data("a", "b"))
  expect_error(bdp_data(1L, 123L))
  expect_error(bdp_data(1L, "b", series_ids = "abc"))
  expect_error(bdp_data(1L, "b", start_date = ""))
  expect_error(bdp_data(1L, "b", end_date = 1L))
  expect_error(bdp_data(1L, "b", lang = "FR"))
  expect_error(bdp_data(1L, "b", updated_after = 1L))
  expect_error(bdp_data(1L, "b", updated_after = TRUE))
  expect_error(bdp_data(1L, "b", updated_after = NA))
})

test_that("bdp_data passes updated_after as obs_published_since", {
  captured <- NULL
  httr2::local_mocked_responses(function(req) {
    captured <<- req
    httr2::response(200L, headers = "content-type: application/json", body = charToRaw("{}"))
  })
  local_mocked_bindings(parse_bdp_data = function(json) data.table())
  bdp_data(54L, "ce3e", updated_after = as.Date("2024-06-01"))
  expect_match(captured$url, "obs_published_since=2024-06-01T00%3A00%3A00")
})

test_that("bdp_series input validation works", {
  expect_error(bdp_series("abc"))
  expect_error(bdp_series(NULL))
  expect_error(bdp_series(1L, lang = "FR"))
})

test_that("bdp_dataset input validation works", {
  expect_error(bdp_dataset("a"))
  expect_error(bdp_dataset(1L, lang = "FR"))
})

test_that("bdp_dimension input validation works", {
  expect_error(bdp_dimension("a"))
  expect_error(bdp_dimension(1L, dimension_id = "a"))
  expect_error(bdp_dimension(1L, lang = "FR"))
})

test_that("bdp_domain input validation works", {
  expect_error(bdp_domain("a"))
  expect_error(bdp_domain(lang = "FR"))
})

test_that("parse_bdp_data works", {
  json <- readRDS(test_path("fixtures", "bdp-data.rds"))
  actual <- parse_bdp_data(json)
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
  expect_numeric(actual$value)
  expect_names(names(actual), must.include = c("date", "value", "freq"))
})

test_that("parse_bdp_series works", {
  json <- readRDS(test_path("fixtures", "bdp-series.rds"))
  actual <- parse_bdp_series(json)
  expect_data_table(actual, nrows = 1L)
  expect_names(
    names(actual),
    must.include = c("id", "label", "dataset_id", "domain_id")
  )
  expect_identical(actual$id, 12518356L)
})

test_that("parse_bdp_dataset works", {
  json <- readRDS(test_path("fixtures", "bdp-dataset.rds"))
  actual <- parse_bdp_dataset(json)
  expect_data_table(actual, min.rows = 1L)
  expect_names(
    names(actual),
    must.include = c("id", "label", "num_series", "obs_updated_at")
  )
})

test_that("parse_bdp_dimension works", {
  items <- readRDS(test_path("fixtures", "bdp-dimension.rds"))
  actual <- parse_bdp_dimension(items)
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("id", "label", "description"))
})

test_that("parse_bdp_category works", {
  json <- readRDS(test_path("fixtures", "bdp-category.rds"))
  actual <- parse_bdp_category(json)
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("id", "label"))
})

test_that("parse_bdp_domain works", {
  json <- readRDS(test_path("fixtures", "bdp-domains.rds"))
  actual <- parse_bdp_domain(json)
  expect_data_table(actual, min.rows = 1L)
  expect_names(
    names(actual),
    must.include = c("id", "label", "has_series", "num_series")
  )
})

#' Fetch National Bank of Poland (NBP) exchange rates
#'
#' Retrieve foreign currency exchange rates from the NBP Web API.
#'
#' @param table (`character(1)`)\cr
#'   Table type: `"a"` (mid rates, major currencies), `"b"` (mid rates, less common currencies),
#'   or `"c"` (bid/ask rates).
#' @param code (`NULL` | `character(1)`)\cr
#'   ISO 4217 currency code (e.g. `"usd"`, `"eur"`). If `NULL`, returns all currencies.
#' @param start_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   Start date of the data.
#' @param end_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   End date of the data.
#' @param last_n (`NULL` | `integer(1)`)\cr
#'   Return only the last `n` quotations.
#' @returns A [data.table::data.table()] with exchange rates.
#' @source <https://api.nbp.pl/en.html>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' nbp_fx_rates("a", "eur")
#' }
nbp_fx_rates <- function(table, code = NULL, start_date = NULL, end_date = NULL, last_n = NULL) {
  assert_choice(table, c("a", "b", "c"))
  assert_string(code, n.chars = 3L, null.ok = TRUE)
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)
  last_n <- assert_count(last_n, positive = TRUE, null.ok = TRUE, coerce = TRUE)
  if (!is.null(last_n) && (!is.null(start_date) || !is.null(end_date))) {
    stop("`last_n` and `start_date`/`end_date` are mutually exclusive.", call. = FALSE)
  }

  resource <- if (is.null(code)) "exchangerates/tables" else "exchangerates/rates"
  path <- nbp_path(resource, table, code, start_date, end_date, last_n)
  json <- nbp(path)
  parse_nbp_fx_rates(json, table, code)
}

#' Fetch National Bank of Poland (NBP) gold prices
#'
#' Retrieve the price of gold from the NBP Web API.
#'
#' @param start_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   Start date of the data.
#' @param end_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   End date of the data.
#' @param last_n (`NULL` | `integer(1)`)\cr
#'   Return only the last `n` quotations.
#' @returns A [data.table::data.table()] with gold prices.
#' @source <https://api.nbp.pl/en.html>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' nbp_gold(last_n = 10L)
#' }
nbp_gold <- function(start_date = NULL, end_date = NULL, last_n = NULL) {
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)
  last_n <- assert_count(last_n, positive = TRUE, null.ok = TRUE, coerce = TRUE)
  if (!is.null(last_n) && (!is.null(start_date) || !is.null(end_date))) {
    stop("`last_n` and `start_date`/`end_date` are mutually exclusive.", call. = FALSE)
  }

  path <- nbp_path("cenyzlota", start_date = start_date, end_date = end_date, last_n = last_n)
  json <- nbp(path)
  parse_nbp_gold(json)
}

parse_nbp_fx_rates <- function(json, table, code) {
  dt <- if (is.null(code)) parse_nbp_tables(json) else parse_nbp_currency(json)
  value_cols <- if (identical(table, "c")) c("bid", "ask") else "mid"
  dt[, c("date", "code", "currency", value_cols), with = FALSE]
}

parse_nbp_tables <- function(json) {
  first <- json[[1L]]
  tables <- if (is.null(first$table)) list(json) else json # nolint
  rbindlist(map(tables, function(tbl) {
    dt <- rbindlist(map(tbl$rates, as.data.table))
    dt[, "date" := as.Date(tbl$effectiveDate)]
  }))
}

parse_nbp_currency <- function(json) {
  dt <- rbindlist(map(json$rates, as.data.table))
  setnames(dt, "effectiveDate", "date")
  dt[, let(date = as.Date(date), code = json$code, currency = json$currency)]
}

parse_nbp_gold <- function(json) {
  data.table(
    date = as.Date(map_chr(json, "data")),
    price = map_dbl(json, "cena")
  )
}

nbp_path <- function(
  resource,
  table = NULL,
  code = NULL,
  start_date = NULL,
  end_date = NULL,
  last_n = NULL
) {
  parts <- c(resource, table, code)
  if (!is.null(last_n)) {
    parts <- c(parts, "last", last_n)
  } else if (!is.null(start_date) && !is.null(end_date)) {
    parts <- c(parts, start_date, end_date)
  } else if (!is.null(start_date)) {
    parts <- c(parts, start_date)
  }
  paste(parts, collapse = "/")
}

nbp <- function(path) {
  request("https://api.nbp.pl/api") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(path) |>
    req_url_query(format = "json") |>
    req_error(body = nbp_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_json()
}

nbp_error_body <- function(resp) {
  resp_body_string(resp)
}

test_that("nbp_fx_rates input validation works", {
  expect_error(nbp_fx_rates("d"))
  expect_error(nbp_fx_rates("a", code = 1L))
  expect_error(nbp_fx_rates("a", start_date = ""))
  expect_error(nbp_fx_rates("a", last_n = -1L))
  expect_error(nbp_fx_rates("a", start_date = "2025-01-01", last_n = 5L), "mutually exclusive")
})

test_that("nbp_gold input validation works", {
  expect_error(nbp_gold(start_date = ""))
  expect_error(nbp_gold(last_n = -1L))
  expect_error(nbp_gold(start_date = "2025-01-01", last_n = 5L), "mutually exclusive")
})

test_that("parse_nbp_fx_rates works for single currency", {
  json <- readRDS(test_path("fixtures", "nbp-fx-rates.rds"))
  actual <- parse_nbp_fx_rates(json, "a", "eur")
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("date", "code", "currency", "mid"))
  expect_date(actual$date)
})

test_that("parse_nbp_fx_rates works for table", {
  json <- readRDS(test_path("fixtures", "nbp-fx-table.rds"))
  actual <- parse_nbp_fx_rates(json, "a", NULL)
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("date", "code", "currency", "mid"))
})

test_that("parse_nbp_fx_rates works for bid/ask", {
  json <- readRDS(test_path("fixtures", "nbp-fx-rates-c.rds"))
  actual <- parse_nbp_fx_rates(json, "c", "eur")
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("date", "code", "currency", "bid", "ask"))
})

test_that("parse_nbp_gold works", {
  json <- readRDS(test_path("fixtures", "nbp-gold.rds"))
  actual <- parse_nbp_gold(json)
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("date", "price"))
  expect_date(actual$date)
})

sdmx_request <- function(base_url, resource, error_body, ..., accept = NULL) {
  request(base_url) |>
    req_user_agent(bbk_user_agent()) |>
    req_headers(accept = accept) |>
    req_url_path_append(resource) |>
    req_url_query(...) |>
    req_error(body = error_body) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_xml()
}

sdmx_data_resource <- function(flow, key, default_key = NULL) {
  flow <- toupper(flow)
  key <- if (is.null(key)) default_key else paste(toupper(key), collapse = "+")
  paste(c("data", flow, key), collapse = "/")
}

sdmx_freq <- function(code) {
  switch(
    code,
    # ISO 8601 duration codes (BBk)
    P1D = "daily",
    P1M = "monthly",
    P3M = "quarterly",
    P1Y = "annual",
    # SDMX frequency codes (ECB, NoB)
    A = "annual",
    S = "semi-annual",
    Q = "quarterly",
    M = "monthly",
    W = "weekly",
    B = ,
    D = "daily"
  )
}

sdmx_dimension <- function(xml, ns_prefix = "str") {
  xpath <- sprintf(".//%s:DimensionList/%s:Dimension", ns_prefix, ns_prefix)
  dims <- xml2::xml_find_all(xml, xpath)
  codelist_xpath <- sprintf(".//%s:Enumeration/Ref", ns_prefix)
  data.table(
    id = xml2::xml_attr(dims, "id"),
    position = as.integer(xml2::xml_attr(dims, "position")),
    codelist = map_chr(dims, \(x) x |> xml2::xml_find_first(codelist_xpath) |> xml2::xml_attr("id"))
  )
}

sdmx_metadata <- function(entries, lang = "en", ns_prefix = "com") {
  xpath <- sprintf(".//%s:Name[@xml:lang='%s']", ns_prefix, lang)
  rbindlist(map(entries, function(node) {
    data.table(
      id = xml2::xml_attr(node, "id"),
      name = node |> xml2::xml_find_all(xpath) |> xml2::xml_text()
    )
  }))
}

test_that("sdmx_data_resource builds correct paths", {
  expect_identical(sdmx_data_resource("exr", NULL), "data/EXR")
  expect_identical(sdmx_data_resource("exr", "d.usd"), "data/EXR/D.USD")
  expect_identical(sdmx_data_resource("exr", c("d.usd", "d.eur")), "data/EXR/D.USD+D.EUR")
  expect_identical(sdmx_data_resource("exr", NULL, default_key = "all"), "data/EXR/all")
})

test_that("sdmx_freq maps codes correctly", {
  expect_identical(sdmx_freq("P1D"), "daily")
  expect_identical(sdmx_freq("P1M"), "monthly")
  expect_identical(sdmx_freq("P3M"), "quarterly")
  expect_identical(sdmx_freq("P1Y"), "annual")
  expect_identical(sdmx_freq("A"), "annual")
  expect_identical(sdmx_freq("S"), "semi-annual")
  expect_identical(sdmx_freq("Q"), "quarterly")
  expect_identical(sdmx_freq("M"), "monthly")
  expect_identical(sdmx_freq("W"), "weekly")
  expect_identical(sdmx_freq("D"), "daily")
  expect_identical(sdmx_freq("B"), "daily")
})

test_that("sdmx_metadata parses entries", {
  body <- xml2::read_xml(test_path("fixtures", "nob-metadata.xml"))
  entries <- xml2::xml_find_all(body, "//str:Dataflow")
  actual <- sdmx_metadata(entries)
  expect_data_table(actual, min.rows = 1L)
  expect_true(all(c("id", "name") %in% names(actual)))
  expect_true("EXR" %in% actual$id)
})

#' Fetch Sveriges Riksbank (SRb) data
#'
#' Retrieve time series data from the Sveriges Riksbank SWEA API.
#'
#' @param series (`character(1)`)\cr
#'   The series ID to query. See [srb_series()] for available series.
#' @param start_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   Start date of the data (e.g., `"2024-01-01"`). If `NULL`, all available data is returned.
#'   Default `NULL`.
#' @param end_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   End date of the data, in the same format as start_date. If `NULL`, data up to the latest
#'   available date is returned. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://developer.api.riksbank.se/>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch USD/SEK exchange rate
#' srb_data("SEKUSDPMI", start_date = "2024-01-01")
#'
#' # fetch EUR/SEK exchange rate
#' srb_data("SEKEURPMI", start_date = "2024-01-01")
#' }
srb_data <- function(series, start_date = NULL, end_date = NULL) {
  assert_string(series, min.chars = 1L)
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)

  if (is.null(start_date) || is.null(end_date)) {
    meta <- srb("Observations/Latest", series)
    start_date <- start_date %??% as.Date("1900-01-01")
    end_date <- end_date %??% as.Date(meta$date)
  }

  json <- srb("Observations", series, format(start_date), format(end_date))
  parse_srb_data(json, series)
}

#' Fetch Sveriges Riksbank (SRb) series metadata
#'
#' Retrieve available series or group metadata from the Sveriges Riksbank SWEA API.
#'
#' @param type (`character(1)`)\cr
#'   The type of metadata to query. One of `"series"` or `"groups"`. Default `"series"`.
#' @returns A [data.table::data.table()] with the requested metadata.
#' @source <https://developer.api.riksbank.se/>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' srb_series()
#' }
srb_series <- function(type = "series") {
  assert_choice(type, c("series", "groups"))
  type <- if (type == "series") "Series" else "Groups"
  json <- srb(type)
  if (type == "series") parse_srb_series(json) else parse_srb_groups(json)
}

parse_srb_data <- function(json, series) {
  if (length(json) == 0L) {
    return(setDT(list(date = as.Date(character()), key = character(), value = numeric())))
  }
  value <- NULL
  dt <- rbindlist(map(json, setDT))
  dt[, let(
    date = as.Date(date),
    key = toupper(series),
    value = as.numeric(value)
  )]
  setcolorder(dt, c("date", "key", "value"))
  dt[]
}

parse_srb_series <- function(json) {
  dt <- json |>
    map(\(x) setDT(map(x, \(v) v %??% NA_character_))) |>
    rbindlist() |>
    setnames(convert_camel_case)
  dt[]
}

parse_srb_groups <- function(json) {
  flatten_groups <- function(groups) {
    res <- list()
    for (g in groups) {
      children <- g$childGroups
      g$childGroups <- NULL
      res <- c(res, list(setDT(map(g, \(v) v %??% NA_character_))))
      if (length(children) > 0L) {
        res <- c(res, flatten_groups(children))
      }
    }
    res
  }

  dt <- json |>
    flatten_groups() |>
    rbindlist() |>
    setnames(convert_camel_case)
  dt[]
}

#' Fetch Sveriges Riksbank (SRb) cross rates
#'
#' Compute cross exchange rates between two currency series from the Sveriges Riksbank SWEA API.
#'
#' @param series1 (`character(1)`)\cr
#'   The first series ID (e.g., `"SEKUSDPMI"`).
#' @param series2 (`character(1)`)\cr
#'   The second series ID (e.g., `"SEKEURPMI"`).
#' @inheritParams srb_data
#' @returns A [data.table::data.table()] with the cross rate data.
#' @source <https://developer.api.riksbank.se/>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # USD/EUR cross rate
#' srb_cross_rates("SEKUSDPMI", "SEKEURPMI", start_date = "2024-01-01", end_date = "2024-01-31")
#' }
srb_cross_rates <- function(series1, series2, start_date, end_date = NULL) {
  assert_string(series1, min.chars = 1L)
  assert_string(series2, min.chars = 1L)
  start_date <- assert_dateish(start_date)
  end_date <- assert_dateish(end_date, null.ok = TRUE)

  args <- list("CrossRates", series1, series2, format(start_date))
  if (!is.null(end_date)) {
    args <- c(args, format(end_date))
  }
  json <- do.call(srb, args)
  series <- paste(series1, series2, sep = "/")
  parse_srb_data(json, series)
}

#' Fetch Sveriges Riksbank (SRb) calendar days
#'
#' Retrieve Swedish banking calendar information from the Sveriges Riksbank SWEA API.
#'
#' @param start_date (`Date(1)` | `character(1)`)\cr
#'   Start date of the range (e.g., `"2024-01-01"`).
#' @param end_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   End date of the range. If `NULL`, data up to the latest available date is returned.
#'   Default `NULL`.
#' @returns A [data.table::data.table()] with the calendar day information.
#' @source <https://developer.api.riksbank.se/>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' srb_calendar("2024-01-01", "2024-01-31")
#' }
srb_calendar <- function(start_date, end_date = NULL) {
  start_date <- assert_dateish(start_date)
  end_date <- assert_dateish(end_date, null.ok = TRUE)

  args <- list("CalendarDays", format(start_date))
  if (!is.null(end_date)) {
    args <- c(args, format(end_date))
  }
  json <- do.call(srb, args)
  parse_srb_calendar(json)
}

parse_srb_calendar <- function(json) {
  dt <- json |>
    map(setDT) |>
    rbindlist() |>
    setnames(convert_camel_case)
  calendar_date <- NULL
  dt[, calendar_date := as.Date(calendar_date)]
  dt[]
}

srb <- function(...) {
  request("https://api.riksbank.se/swea/v1") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(...) |>
    req_error(body = srb_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_json()
}

srb_error_body <- function(resp) {
  resp_body_string(resp, "UTF-8")
}

test_that("srb_data input validation works", {
  expect_error(srb_data(1L))
  expect_error(srb_data(TRUE))
  expect_error(srb_data(NULL))
  expect_error(srb_data(NA))
  expect_error(srb_data(c("a", "b")))
  # start_date
  expect_error(srb_data("SEKUSDPMI", start_date = TRUE))
  expect_error(srb_data("SEKUSDPMI", start_date = c("a", "b")))
  # end_date
  expect_error(srb_data("SEKUSDPMI", end_date = TRUE))
  expect_error(srb_data("SEKUSDPMI", end_date = c("a", "b")))
})

test_that("parse_srb_data works", {
  json <- jsonlite::fromJSON(test_path("fixtures", "srb-data.json"), simplifyVector = FALSE)
  actual <- parse_srb_data(json, "SEKUSDPMI")
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
  expect_numeric(actual$value)
  expect_true(all(c("date", "key", "value") %in% names(actual)))
})

test_that("parse_srb_data handles empty response", {
  actual <- parse_srb_data(list(), "SEKUSDPMI")
  expect_data_table(actual, nrows = 0L)
  expect_true(all(c("date", "key", "value") %in% names(actual)))
})

test_that("srb_cross_rates input validation works", {
  expect_error(srb_cross_rates(1L, "b", "2024-01-01"))
  expect_error(srb_cross_rates("a", 1L, "2024-01-01"))
  expect_error(srb_cross_rates("a", "b", NULL))
  expect_error(srb_cross_rates("a", "b", "2024-01-01", end_date = TRUE))
})

test_that("srb_cross_rates works", {
  json <- readRDS(test_path("fixtures", "srb-crossrates.rds"))
  actual <- parse_srb_data(json, "SEKUSDPMI/SEKEURPMI")
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
  expect_numeric(actual$value)
  expect_identical(unique(actual$key), "SEKUSDPMI/SEKEURPMI")
})

test_that("srb_calendar input validation works", {
  expect_error(srb_calendar(1L))
  expect_error(srb_calendar(NULL))
  expect_error(srb_calendar("2024-01-01", end_date = TRUE))
})

test_that("parse_srb_calendar works", {
  json <- readRDS(test_path("fixtures", "srb-calendar.rds"))
  actual <- parse_srb_calendar(json)
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$calendar_date)
  expect_logical(actual$swedish_bankday)
  expect_names(names(actual), must.include = c("calendar_date", "swedish_bankday", "week_number"))
})

test_that("srb_series input validation works", {
  expect_error(srb_series(1L))
  expect_error(srb_series("invalid"))
})

test_that("parse_srb_series works", {
  json <- jsonlite::fromJSON(test_path("fixtures", "srb-series.json"), simplifyVector = FALSE)
  actual <- parse_srb_series(json)
  expect_data_table(actual, min.rows = 1L)
  expect_true(all(c("series_id", "source", "long_description") %in% names(actual)))
  expect_true("SEKUSDPMI" %in% actual$series_id)
})

test_that("parse_srb_groups works", {
  json <- jsonlite::fromJSON(test_path("fixtures", "srb-groups.json"), simplifyVector = FALSE)
  actual <- parse_srb_groups(json)
  expect_data_table(actual, min.rows = 1L)
  expect_true(all(c("group_id", "name") %in% names(actual)))
})

map <- function(.x, .f, ...) {
  if (is.function(.f)) {
    lapply(.x, .f, ...)
  } else {
    lapply(.x, `[[`, .f, ...)
  }
}

map_mold <- function(.x, .f, .value, ...) {
  out <- if (is.function(.f)) {
    vapply(.x, .f, FUN.VALUE = .value, USE.NAMES = FALSE, ...)
  } else {
    vapply(.x, `[[`, .f, FUN.VALUE = .value, USE.NAMES = FALSE, ...)
  }
  stats::setNames(out, names(.x))
}

map_lgl <- function(.x, .f, ...) {
  map_mold(.x, .f, NA, ...)
}

map_int <- function(.x, .f, ...) {
  map_mold(.x, .f, NA_integer_, ...)
}

map_dbl <- function(.x, .f, ...) {
  map_mold(.x, .f, NA_real_, ...)
}

map_chr <- function(.x, .f, ...) {
  map_mold(.x, .f, NA_character_, ...)
}

#' Fetch Bank of Canada (BoC) data
#'
#' Retrieve time series data from the Bank of Canada Valet API.
#'
#' @param group_name (`NULL` | `character(1)`)\cr
#'   Name of the group. Only one of `group_name` or `series_name` can be used.
#' @param series_name (`NULL` | `character()`)\cr
#'   Name of the series.
#' @param start_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   Start date of the data. Default `NULL`.
#' @param end_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   End date of the data. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.bankofcanada.ca/valet/docs>
#' @family data
#' @export
#' @examples
#' \dontrun{
#' # fetch all data for a single group
#' dt <- boc_data(group_name = "FX_RATES_DAILY")
#' head(dt)
#'
#' # or for multiple series ids
#' dt <- boc_data(
#'   series_name = c("FXUSDCAD", "FXEURCAD"),
#'   start_date = "2023-01-23",
#'   end_date = "2023-07-19"
#' )
#' head(dt)
#' }
boc_data <- function(group_name = NULL, series_name = NULL, start_date = NULL, end_date = NULL) {
  assert_string(group_name, min.chars = 1L, null.ok = TRUE)
  assert_character(series_name, min.chars = 1L, min.len = 1L, null.ok = TRUE)
  if (!is.null(group_name) && !is.null(series_name)) {
    stop("Only one of `group_name` or `series_name` must be provided.", call. = FALSE)
  }
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)
  if (!is.null(group_name)) {
    boc_group_obs(group_name, start_date, end_date)
  } else {
    boc_series_obs(series_name, start_date, end_date)
  }
}

#' Fetch Bank of Canada (BoC) metadata (details)
#'
#' @inheritParams boc_data
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.bankofcanada.ca/valet/docs>
#' @export
#' @examples
#' \dontrun{
#' boc_metadata(group_name = "FX_RATES_DAILY")
#' boc_metadata(series_name = "FXUSDCAD")
#' }
boc_metadata <- function(group_name = NULL, series_name = NULL) {
  assert_string(group_name, min.chars = 1L, null.ok = TRUE)
  assert_string(series_name, min.chars = 1L, null.ok = TRUE)
  if (!is.null(group_name) && !is.null(series_name)) {
    stop("Only one of `group_name` or `series_name` must be provided.", call. = FALSE)
  }
  if (!is.null(group_name)) boc_details_group(group_name) else boc_details_series(series_name)
}

#' Fetch Bank of Canada (BoC) available series or group
#'
#' Access all available series or groups from the Bank of Canada Valet API.
#'
#' @param type (`character(1)`)\cr
#'   Set of data to return. One of `"groups"` or `"series"`. Default `"groups"`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.bankofcanada.ca/valet/docs>
#' @export
#' @examples
#' \dontrun{
#' catalog <- boc_catalog()
#' head(catalog)
#'
#' # filter for effective exchange rate series
#' dt <- catalog[grepl("CEER", label)]
#' head(dt)
#' }
boc_catalog <- function(type = "groups") {
  assert_choice(type, c("groups", "series"))
  json <- boc("lists", type)
  lst <- json[[type]]
  dt <- rbindlist(map(lst, setDT))
  dt[, "name" := names(lst)]
  dt[]
}

boc_details_series <- function(name) {
  json <- boc("series", name)
  dt <- setDT(json$seriesDetails)
  dt
}

boc_details_group <- function(name) {
  json <- boc("groups", name)
  grp <- json$groupDetails
  meta <- setDT(grp[lengths(grp) == 1L])
  setnames(meta, \(x) paste("group", x, sep = "_"))
  series <- rbindlist(map(grp$groupSeries, setDT))
  series[, "name" := names(grp$groupSeries)]
  setnames(series, \(x) paste("series", x, sep = "_"))
  series[, names(meta) := meta]
  cols <- names(series)
  setcolorder(series, cols[startsWith(cols, "group_")])
  series[]
}

boc_series_obs <- function(name, start_date, end_date) {
  name <- paste(name, collapse = ",")
  json <- boc("observations", name, start_date = start_date, end_date = end_date)

  meta <- rbindlist(map(json$seriesDetail, \(x) setDT(x[lengths(x) == 1])))
  meta[, "name" := names(json$seriesDetail)]

  obs <- json$observations |>
    map(function(x) {
      nms <- names(x)
      x |>
        unlist(recursive = FALSE, use.names = FALSE) |>
        setDT() |>
        setnames(nms)
    }) |>
    rbindlist(fill = TRUE) |>
    setnames("d", "date") |>
    melt(
      id.vars = "date",
      variable.name = "name",
      value.name = "value",
      na.rm = TRUE,
      variable.factor = FALSE
    )

  obs <- meta[obs, on = "name"]
  value <- NULL
  obs[, let(date = as.Date(date), value = as.numeric(value))]
  obs[]
}

boc_group_obs <- function(name = "FX_RATES_DAILY", start_date = NULL, end_date = NULL) {
  json <- boc("observations/group", name)
  grp <- setDT(map(json$groupDetail, \(x) x %??% NA_character_))
  setnames(grp, \(x) paste("group", x, sep = "_"))
  meta <- rbindlist(map(json$seriesDetail, \(x) setDT(x[lengths(x) == 1])))
  meta[, "name" := names(json$seriesDetail)]
  setnames(meta, \(x) paste("series", x, sep = "_"))

  obs <- json$observations |>
    map(function(x) {
      nms <- names(x)
      x |>
        unlist(recursive = FALSE, use.names = FALSE) |>
        setDT() |>
        setnames(nms)
    }) |>
    rbindlist(fill = TRUE) |>
    setnames("d", "date") |>
    melt(
      id.vars = "date",
      variable.name = "name",
      value.name = "value",
      na.rm = TRUE,
      variable.factor = FALSE
    )
  value <- NULL
  obs[, let(date = as.Date(date), value = as.numeric(value))]
  setnames(obs, "name", "series_name")

  obs <- meta[obs, on = "series_name"]
  obs[, names(grp) := grp]
  obs[]
}

boc <- function(resource, name, ...) {
  request("https://www.bankofcanada.ca/valet") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(resource, name, "json") |>
    req_url_query(...) |>
    req_error(body = boc_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_json()
}

boc_error_body <- function(resp) {
  content_type <- resp_content_type(resp)
  if (identical(content_type, "application/json")) {
    json <- resp_body_json(resp)
    msg <- json$message
    docs <- sprintf("See docs at <%s>", json$docs)
    c(msg, docs)
  }
}

check_dateish <- function(x, null.ok = FALSE) {
  if (null.ok && is.null(x)) {
    return(TRUE)
  }
  if (test_date(x, any.missing = FALSE, len = 1L)) {
    return(TRUE)
  }
  if (test_string(x, pattern = "^\\d{4}-\\d{2}-\\d{2}$")) {
    return(TRUE)
  }
  if (null.ok) {
    "Must be `NULL`, a Date, or a string in 'YYYY-MM-DD' format"
  } else {
    "Must be a Date or a string in 'YYYY-MM-DD' format"
  }
}

assert_dateish <- function(x, null.ok = FALSE, .var.name = vname(x), add = NULL) {
  res <- check_dateish(x, null.ok)
  makeAssertion(x, res, .var.name, add)
  if (!is.null(x) && !inherits(x, "Date")) {
    x <- as.Date(x)
  }
  invisible(x)
}

check_timestampish <- function(x, null.ok = FALSE) {
  if (null.ok && is.null(x)) {
    return(TRUE)
  }
  if (test_multi_class(x, c("POSIXt", "Date")) && length(x) == 1L && !anyNA(x)) {
    return(TRUE)
  }
  if (test_string(x, min.chars = 4L)) {
    return(TRUE)
  }
  if (null.ok) {
    "Must be `NULL`, a Date, a POSIXct, or an ISO 8601 string"
  } else {
    "Must be a Date, a POSIXct, or an ISO 8601 string"
  }
}

assert_timestampish <- function(x, null.ok = FALSE, .var.name = vname(x), add = NULL) {
  res <- check_timestampish(x, null.ok)
  makeAssertion(x, res, .var.name, add)
  if (is.null(x)) {
    return(invisible())
  }
  if (inherits(x, c("POSIXt", "Date"))) {
    x <- format(as.POSIXct(x, tz = "UTC"), "%Y-%m-%dT%H:%M:%S", tz = "UTC")
  }
  invisible(x)
}

assert_period <- function(x, .var.name = vname(x)) {
  assert(
    check_null(x),
    check_string(x, min.chars = 1L),
    check_count(x, positive = TRUE),
    .var.name = .var.name
  )
  invisible(x)
}

#' Fetch Bank of Japan (BoJ) data
#'
#' Retrieve time series data from the Bank of Japan Statistics API.
#'
#' @param db (`character(1)`)\cr
#'   The database code to query (e.g., `"FM08"` for foreign exchange rates). See the
#'   [API manual](https://www.stat-search.boj.or.jp/info/api_manual_en.pdf) for available databases.
#' @param code (`character()`)\cr
#'   One or more series codes to query (e.g., `"FXERD01"` for USD/JPY spot rate). Maximum 250
#'   codes per request. All codes must have the same frequency. Use [boj_metadata()] to find
#'   available codes.
#' @param start_date (`NULL` | `character(1)` | `integer(1)`)\cr
#'   Start date of the data. Format depends on frequency: `"YYYYMMDD"` or `YYYY` for daily,
#'   `"YYYYMM"` for monthly, `"YYYYQQ"` for quarterly (where QQ is 01-04), `"YYYY"` for annual.
#'   If `NULL`, all available data is returned. Default `NULL`.
#' @param end_date (`NULL` | `character(1)` | `integer(1)`)\cr
#'   End date of the data, in the same format as start_date. If `NULL`, data up to the latest
#'   available date is returned. Default `NULL`.
#' @param lang (`character(1)`)\cr
#'   Language for series names, either `"en"` or `"jp"`. Default `"en"`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.stat-search.boj.or.jp/index_en.html>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch USD/JPY exchange rate
#' boj_data("FM08", "FXERD01", start_date = "202401")
#'
#' # fetch multiple exchange rates
#' boj_data("FM08", c("FXERD01", "FXERD02"), start_date = "202401")
#' }
boj_data <- function(db, code, start_date = NULL, end_date = NULL, lang = "en") {
  assert_string(db, min.chars = 1L)
  assert_character(code, min.chars = 1L, min.len = 1L, max.len = 250L)
  assert_boj_date(start_date)
  assert_boj_date(end_date)
  assert_choice(lang, c("en", "jp"))

  req <- boj_request(
    "getDataCode",
    db = toupper(db),
    code = code,
    startDate = start_date,
    endDate = end_date,
    lang = lang
  )
  resps <- req_perform_iterative(
    req,
    next_req = iterate_with_cursor("startPosition", boj_next_position),
    max_reqs = Inf
  )
  series <- resps_data(resps, \(resp) resp_body_json(resp)$RESULTSET)
  parse_boj_data(list(RESULTSET = series))
}

boj_next_position <- function(resp) {
  pos <- resp_body_json(resp)$NEXTPOSITION
  if (length(pos) == 0L) NULL else pos
}

#' Fetch Bank of Japan (BoJ) metadata
#'
#' Retrieve series metadata from the Bank of Japan Statistics API.
#'
#' @param db (`character(1)`)\cr
#'   The database code to query (e.g., `"FM08"` for foreign exchange rates).
#' @param lang (`character(1)`)\cr
#'   Language for names, either `"en"` or `"jp"`. Default `"en"`.
#' @returns A [data.table::data.table()] with the requested metadata.
#' @source <https://www.stat-search.boj.or.jp/index_en.html>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' boj_metadata("FM08")
#' }
boj_metadata <- function(db, lang = "en") {
  assert_string(db, min.chars = 1L)
  assert_choice(lang, c("en", "jp"))

  json <- boj("getMetadata", db = toupper(db), lang = lang)
  parse_boj_metadata(json)
}

parse_boj_data <- function(json) {
  series <- json$RESULTSET
  if (length(series) == 0L) {
    return(setDT(list(
      date = as.Date(character()),
      key = character(),
      value = numeric(),
      freq = character(),
      name = character(),
      unit = character()
    )))
  }

  dt <- rbindlist(map(series, function(x) {
    dates <- unlist(x$VALUES$SURVEY_DATES)
    values <- as.numeric(map(x$VALUES$VALUES, \(x) x %??% NA_real_))
    freq <- tolower(x$FREQUENCY)

    dt <- as.data.table(list(
      date = parse_boj_date(dates, freq),
      key = x$SERIES_CODE,
      value = values,
      freq = boj_freq(freq),
      name = x$NAME_OF_TIME_SERIES %??% x$NAME_OF_TIME_SERIES_J,
      unit = x$UNIT %??% x$UNIT_J
    ))
    na.omit(dt, cols = "value")
  }))
  dt[]
}

parse_boj_metadata <- function(json) {
  series <- json$RESULTSET
  if (length(series) == 0L) {
    return(data.table(
      code = character(),
      name = character(),
      unit = character(),
      frequency = character(),
      category = character()
    ))
  }

  dt <- rbindlist(map(series, function(x) {
    code <- x$SERIES_CODE
    if (is.null(code) || !nzchar(code)) {
      return()
    }
    data.table(
      code = code,
      name = x$NAME_OF_TIME_SERIES %??% x$NAME_OF_TIME_SERIES_J %??% NA_character_,
      unit = x$UNIT %??% x$UNIT_J %??% NA_character_,
      frequency = x$FREQUENCY %??% NA_character_,
      category = x$CATEGORY %??% x$CATEGORY_J %??% NA_character_
    )
  }))
  dt[]
}

parse_boj_date <- function(dates, freq) {
  switch(
    freq,
    daily = ,
    `weekly(monday)` = ,
    `weekly(tuesday)` = ,
    `weekly(wednesday)` = ,
    `weekly(thursday)` = ,
    `weekly(friday)` = ,
    `weekly(saturday)` = ,
    `weekly(sunday)` = {
      as.Date(as.character(dates), format = "%Y%m%d")
    },
    monthly = as.Date(paste0(dates, "01"), format = "%Y%m%d"),
    quarterly = {
      dates_chr <- as.character(dates)
      year <- substr(dates_chr, 1L, 4L)
      quarter <- as.integer(substr(dates_chr, 5L, 6L))
      month <- sprintf("%02d", (quarter - 1L) * 3L + 1L)
      as.Date(paste0(year, month, "01"), format = "%Y%m%d")
    },
    semiannual = ,
    `semiannual(sep)` = {
      dates_chr <- as.character(dates)
      year <- substr(dates_chr, 1L, 4L)
      half <- as.integer(substr(dates_chr, 5L, 6L))
      month <- sprintf("%02d", (half - 1L) * 6L + 1L)
      as.Date(paste0(year, month, "01"), format = "%Y%m%d")
    },
    annual = ,
    `annual(mar)` = as.integer(dates),
    as.character(dates)
  )
}

boj_freq <- function(freq) {
  switch(
    freq,
    daily = "daily",
    monthly = "monthly",
    quarterly = "quarterly",
    semiannual = ,
    `semiannual(sep)` = "semi-annual",
    annual = ,
    `annual(mar)` = "annual",
    `weekly(monday)` = ,
    `weekly(tuesday)` = ,
    `weekly(wednesday)` = ,
    `weekly(thursday)` = ,
    `weekly(friday)` = ,
    `weekly(saturday)` = ,
    `weekly(sunday)` = "weekly",
    freq
  )
}

assert_boj_date <- function(x, .var.name = vname(x)) {
  assert(
    check_null(x),
    check_string(x, min.chars = 4L, pattern = "^\\d+$"),
    check_count(x, positive = TRUE),
    .var.name = .var.name
  )
  invisible(x)
}

boj_error_body <- function(resp) {
  json <- resp_body_json(resp)
  json$MESSAGE
}

boj_request <- function(resource, ...) {
  request("https://www.stat-search.boj.or.jp/api/v1") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(resource) |>
    req_url_query(..., .multi = "comma") |>
    req_error(body = boj_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache()
}

boj <- function(resource, ...) {
  boj_request(resource, ...) |>
    req_perform() |>
    resp_body_json()
}

test_that("boj_data input validation works", {
  expect_error(boj_data(1L, "FXERD01"))
  expect_error(boj_data(TRUE, "FXERD01"))
  expect_error(boj_data(NULL, "FXERD01"))
  expect_error(boj_data(NA, "FXERD01"))
  # code
  expect_error(boj_data("FM08", 1L))
  expect_error(boj_data("FM08", TRUE))
  expect_error(boj_data("FM08", NULL))
  # start_date
  expect_error(boj_data("FM08", "FXERD01", start_date = TRUE))
  expect_error(boj_data("FM08", "FXERD01", start_date = c("a", "b")))
  # end_date
  expect_error(boj_data("FM08", "FXERD01", end_date = TRUE))
  expect_error(boj_data("FM08", "FXERD01", end_date = c("a", "b")))
  # lang
  expect_error(boj_data("FM08", "FXERD01", lang = "fr"))
})

test_that("parse_boj_data works", {
  json <- jsonlite::fromJSON(test_path("fixtures", "boj-data.json"), simplifyVector = FALSE)
  actual <- parse_boj_data(json)
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
  expect_numeric(actual$value)
  expect_true(all(c("date", "key", "value", "freq") %in% names(actual)))
  expect_true("FXERD01" %in% actual$key)
  expect_identical(unique(actual$freq), "daily")
})

test_that("parse_boj_data handles empty response", {
  json <- list(RESULTSET = list())
  actual <- parse_boj_data(json)
  expect_data_table(actual, nrows = 0L)
  expect_true(all(c("date", "key", "value") %in% names(actual)))
})

test_that("boj_data follows NEXTPOSITION across pages", {
  json_resp <- function(body) {
    httr2::response_json(body = body)
  }
  series <- function(code) {
    list(
      SERIES_CODE = code,
      NAME_OF_TIME_SERIES = code,
      UNIT = "x",
      FREQUENCY = "DAILY",
      VALUES = list(SURVEY_DATES = list("20240101"), VALUES = list("1"))
    )
  }
  httr2::local_mocked_responses(list(
    json_resp(list(NEXTPOSITION = 2L, RESULTSET = list(series("A")))),
    json_resp(list(NEXTPOSITION = NULL, RESULTSET = list(series("B"), series("C"))))
  ))

  x <- boj_data("FM08", c("A", "B", "C"))
  expect_identical(sort(unique(x$key)), c("A", "B", "C"))
  expect_identical(nrow(x), 3L)
})

test_that("boj_metadata input validation works", {
  expect_error(boj_metadata(1L))
  expect_error(boj_metadata(TRUE))
  expect_error(boj_metadata(NULL))
  expect_error(boj_metadata(NA))
  expect_error(boj_metadata("FM08", lang = "fr"))
})

test_that("parse_boj_metadata works", {
  json <- jsonlite::fromJSON(test_path("fixtures", "boj-metadata.json"), simplifyVector = FALSE)
  actual <- parse_boj_metadata(json)
  expect_data_table(actual, min.rows = 1L)
  expect_true(all(c("code", "name", "unit", "frequency", "category") %in% names(actual)))
  expect_true("FXERD01" %in% actual$code)
  expect_true(all(nzchar(actual$code)))
})

test_that("parse_boj_metadata handles empty response", {
  json <- list(RESULTSET = list())
  actual <- parse_boj_metadata(json)
  expect_data_table(actual, nrows = 0L)
})

test_that("parse_boj_date handles different frequencies", {
  expect_identical(parse_boj_date(20250115, "daily"), as.Date("2025-01-15"))
  expect_identical(parse_boj_date(202501, "monthly"), as.Date("2025-01-01"))
  expect_identical(parse_boj_date(202501, "quarterly"), as.Date("2025-01-01"))
  expect_identical(parse_boj_date(202503, "quarterly"), as.Date("2025-07-01"))
  expect_identical(parse_boj_date(2025, "annual"), 2025L)
})

test_that("boj_freq maps correctly", {
  expect_identical(boj_freq("daily"), "daily")
  expect_identical(boj_freq("monthly"), "monthly")
  expect_identical(boj_freq("quarterly"), "quarterly")
  expect_identical(boj_freq("annual"), "annual")
  expect_identical(boj_freq("annual(mar)"), "annual")
  expect_identical(boj_freq("semiannual"), "semi-annual")
  expect_identical(boj_freq("weekly(monday)"), "weekly")
})

test_that("assert_boj_date works", {
  expect_silent(assert_boj_date(NULL))
  expect_silent(assert_boj_date("202501"))
  expect_silent(assert_boj_date("2025"))
  expect_silent(assert_boj_date(2025L))
  expect_error(assert_boj_date(TRUE))
  expect_error(assert_boj_date("abc"))
  expect_error(assert_boj_date(c("a", "b")))
})

#' Fetch Swiss National Bank (SNB) data
#'
#' Retrieve time series data from the SNB data portal.
#'
#' @param key (`character(1)`)\cr
#'   The series key to query.
#' @param start_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   Start date of the data.
#' @param end_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   End date of the data.
#' @param lang (`character(1)`)\cr
#'   Language to query, either `"en"` or `"de"`. Default `"en"`.
#' @returns A [data.table::data.table()] with the requested data.
#' @export
#' @source <https://data.snb.ch/en>
#' @family data
#' @examplesIf curl::has_internet()
#' \donttest{
#' snb_data("rendopar")
#'
#' # or filter for date range
#' snb_data("rendopar", "2020-01-01", "2020-12-31")
#' }
snb_data <- function(key, start_date = NULL, end_date = NULL, lang = "en") {
  assert_string(key, min.chars = 1L)
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)
  assert_choice(lang, c("en", "de"))

  json <- snb(id = key, fromDate = start_date, toDate = end_date, lang = lang)
  parse_snb_data(json)
}

parse_snb_data <- function(json) {
  dt <- rbindlist(map(json$timeseries, function(x) {
    meta <- as.data.table(x$metadata)
    header <- x$header
    cols <- map_chr(header, "dim")
    cols <- gsub("[[:space:][:punct:]]", "_", tolower(cols))
    item <- setNames(map(header, "dimItem"), cols)
    ref <- setDT(item)
    vals <- x$values
    vals <- data.table(
      date = map_chr(vals, "date"),
      value = map_dbl(vals, "value")
    )
    vals[, names(meta) := meta]
    vals[, names(ref) := ref]
  }))

  dt[!nzchar(scale), scale := NA_character_]
  setnames(dt, "frequency", "duration")
  duration <- NULL
  dt[, duration := substring(duration, 1L, 3L)]
  freq <- switch(
    dt[1L, duration],
    P1M = "monthly",
    P3M = "quarterly",
    P1Y = "annual",
    P1D = "daily"
  )
  dt[, let(date = parse_date(date, freq), freq = freq)]
  setcolorder(dt, col_order, skip_absent = TRUE)
  dt[]
}

#' Fetch Swiss National Bank (SNB) dimensions
#'
#' Retrieve the dimension structure for a given cube from the SNB data portal.
#'
#' @inheritParams snb_data
#' @returns A [data.table::data.table()] with the dimension structure.
#' @export
#' @source <https://data.snb.ch/en>
#' @family metadata
#' @examplesIf curl::has_internet()
#' \donttest{
#' snb_dimension("rendopar")
#' }
snb_dimension <- function(key, lang = "en") {
  assert_string(key, min.chars = 1L)
  assert_choice(lang, c("en", "de"))

  json <- snb(key, resource = "dimensions", lang = lang)
  parse_snb_dimension(json)
}

parse_snb_dimension <- function(json) {
  rbindlist(map(json$dimensions, function(x) {
    items <- x$dimensionItems
    has_children <- map_lgl(items, \(item) !is.null(item$dimensionItems))
    if (any(has_children)) {
      items <- unlist(map(items, "dimensionItems"), recursive = FALSE)
    }
    data.table(
      dim_id = x$id,
      dim_name = x$name,
      item_id = map_chr(items, "id"),
      item_name = map_chr(items, "name")
    )
  }))
}

snb <- function(id, ..., resource = "data/json", lang = "en") {
  url <- sprintf("https://data.snb.ch/api/cube/%s/%s", id, resource)
  request(url) |>
    req_user_agent(bbk_user_agent()) |>
    req_url_path_append(lang) |>
    req_url_query(...) |>
    req_error(body = snb_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_json(check_type = FALSE)
}

snb_error_body <- function(resp) {
  body <- resp_body_string(resp, "UTF-8")
  msg <- jsonlite::fromJSON(body)
  docs <- "See docs at <https://data.snb.ch/en>"
  c(msg$message, docs)
}

test_that("snb_data input validation works", {
  expect_error(snb_data(c("id1", "id2")))
  expect_error(snb_data(123L))
  expect_error(snb_data(NULL))
  expect_error(snb_data(NA))
  # start_date should be a valid date
  expect_error(snb_data("abc", ""))
  expect_error(snb_data("abc", 1L))
  expect_error(snb_data("abc", "2020"))
  expect_error(snb_data("abc", end_date = ""))
  expect_error(snb_data("abc", end_date = 1L))
  expect_error(snb_data("abc", end_date = "2020"))
  # lang should be one of "en", "de"
  expect_error(snb_data("abc", lang = "fr"))
})

test_that("snb_dimension input validation works", {
  expect_error(snb_dimension(c("id1", "id2")))
  expect_error(snb_dimension(123L))
  expect_error(snb_dimension(NULL))
  expect_error(snb_dimension("abc", lang = "fr"))
})

test_that("parse_snb_dimension() works", {
  body <- readRDS(test_path("fixtures", "snb-dimensions.rds"))
  actual <- parse_snb_dimension(body)
  expect_data_table(actual, min.rows = 1L)
  expect_names(names(actual), must.include = c("dim_id", "dim_name", "item_id", "item_name"))
})

test_that("parse_snb_data() works", {
  body <- readRDS(test_path("fixtures", "snb-data.rds"))
  actual <- parse_snb_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_match(actual$key, "^EPB@SNB.rendopar")
  expect_date(actual$date)
})

#' Get or manage the bbk API cache
#'
#' `bbk_cache_dir()` returns the path where cached API responses are stored.
#' `bbk_cache_clear()` clears all cached responses.
#'
#' @details
#' The cache is only used when enabled with `options(bbk.cache = TRUE)`.
#' Cached responses are stored for 1 day by default, but this can be customized with
#' `options(bbk.cache_max_age = seconds)`.
#'
#' @name cache
#' @examples
#' \dontrun{
#' # enable caching
#' options(bbk.cache = TRUE)
#'
#' # view cache location
#' bbk_cache_dir()
#'
#' # clear the cache
#' bbk_cache_clear()
#' }
NULL

#' @rdname cache
#' @export
bbk_cache_dir <- function() {
  file.path(tools::R_user_dir("bbk", "cache"), "httr2")
}

#' @rdname cache
#' @export
bbk_cache_clear <- function() {
  cache_dir <- bbk_cache_dir()
  if (dir.exists(cache_dir)) {
    unlink(dir(cache_dir, full.names = TRUE))
  }
}

req_bbk_retry <- function(req) {
  req_retry(req, max_tries = 3L)
}

req_bbk_cache <- function(req) {
  if (isTRUE(getOption("bbk.cache", FALSE))) {
    req <- req_cache(
      req,
      path = bbk_cache_dir(),
      max_age = getOption("bbk.cache_max_age", 86400L) # 1 day
    )
  }
  req
}

grepv <- function(...) {
  grep(..., value = TRUE)
}

parse_date <- function(date, freq) {
  switch(
    freq,
    daily = as.Date(date),
    monthly = as.Date(paste0(date, "-01")),
    annual = as.integer(date),
    date
  )
}

extract_metadata <- function(string, pattern, fixed = FALSE) {
  x <- grepv(pattern, string, fixed = fixed)
  if (length(x) > 0L) {
    strsplit(x, ",", fixed = TRUE)[[1L]][[2L]]
  } else {
    NA_character_
  }
}

convert_camel_case <- function(x) {
  tolower(gsub("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))", "_\\1", x, perl = TRUE))
}

#' Fetch Bank of England (BoE) data
#'
#' Retrieve time series data from the BoE database.
#'
#' @param key (`character()`)\cr
#'   The series keys to query.
#' @param start_date (`character(1)` | `Date(1)`)\cr
#'   Start date of the data.
#' @param end_date (`character(1)` | `Date(1)`)\cr
#'   End date of the data. Default is today's date.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.bankofengland.co.uk/boeapps/database>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # Bank Rate
#' boe_data("IUDBEDR", "2015-01-01")
#'
#' # SONIA daily rate
#' boe_data("IUDSOIA", "2015-01-01")
#'
#' # 10-year nominal par yield
#' boe_data("IUDMNPY", "2015-01-01")
#'
#' # multiple series
#' boe_data(c("IUMABEDR", "IUALBEDR"), "2015-01-01")
#' }
boe_data <- function(key, start_date, end_date = Sys.Date()) {
  assert_character(key, min.chars = 1L, max.len = 300L)
  start_date <- assert_dateish(start_date)
  end_date <- assert_dateish(end_date)

  xml <- boe(
    SeriesCodes = key,
    Datefrom = format(start_date, "%d/%b/%Y"),
    Dateto = format(end_date, "%d/%b/%Y")
  )
  parse_boe_data(xml)
}

parse_boe_data <- function(xml) {
  xml |>
    xml2::xml_ns_strip() |>
    xml2::xml_children() |>
    map(function(x) {
      id <- xml2::xml_attr(x, "SCODE")
      desc <- xml2::xml_attr(x, "DESC")
      freq_name <- x |>
        xml2::xml_find_first("./Cube[@FREQ_NAME]") |>
        xml2::xml_attr("FREQ_NAME") |>
        tolower()
      attrs <- xml2::xml_find_all(x, "./Cube[@CAT_NAME]")
      nms <- xml2::xml_attr(attrs, "CAT_NAME")
      nms <- chartr(" ", "_", tolower(nms))
      attrs <- attrs |>
        xml2::xml_attr("VAL_DESC") |>
        setNames(nms) |>
        as.list()

      vals <- xml2::xml_find_all(x, "./Cube[@TIME and @OBS_VALUE]")
      date <- as.Date(xml2::xml_attr(vals, "TIME"))
      value <- as.numeric(xml2::xml_attr(vals, "OBS_VALUE"))
      dt <- data.table(date = date, id = id, value = value, description = desc, freq = freq_name)
      setnames(dt, "id", "key")
      dt[, (names(attrs)) := attrs]
    }) |>
    rbindlist(fill = TRUE)
}

boe <- function(...) {
  request("https://www.bankofengland.co.uk/boeapps/database/_iadb-fromshowcolumns.asp") |>
    req_user_agent(bbk_user_agent()) |>
    req_url_query(..., xml.x = "yes", .multi = "comma") |>
    req_error(
      is_error = function(resp) {
        resp_status(resp) >= 400L || grepl("errorpage", resp_url(resp), ignore.case = TRUE)
      },
      body = boe_error_body
    ) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_xml()
}

boe_error_body <- function(resp) {
  msg <- "The BoE returned an error for the request."
  c(msg, sprintf("See docs at <%s>", resp_url(resp)))
}

test_that("boe_data input validation works", {
  expect_error(boe_data(c("id1", "id2")))
  expect_error(boe_data(123L))
  expect_error(boe_data(NULL))
  expect_error(boe_data(NA))
  # start_date should be a valid date
  expect_error(boe_data("abc", ""))
  expect_error(boe_data("abc", 1L))
  expect_error(boe_data("abc", "2020"))
  expect_error(boe_data("abc", end_date = ""))
  expect_error(boe_data("abc", end_date = 1L))
  expect_error(boe_data("abc", end_date = "2020"))
})

test_that("parse_boe_data() works", {
  body <- xml2::read_xml(test_path("fixtures", "boe-data.xml"))
  actual <- parse_boe_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
})

#' Fetch Euro foreign exchange reference rates
#'
#' Fetch the latest or historical Euro foreign exchange reference rates from the
#' European Central Bank (ECB).
#'
#' @details
#' Note you can achieve the same by calling the [ecb_data()] function with the right
#' parameters for each currency.
#'
#' The reference rates are usually updated at around 16:00 CET every working day, except on [TARGET
#' closing days](https://www.ecb.europa.eu/ecb/contacts/working-hours/html/index.en.html).
#'
#' They are based on the daily concertation procedure between central banks across
#' Europe, which normally takes place around 14:10 CET. The reference rates are
#' published for information purposes only.
#' Using the rates for transaction purposes is strongly discouraged.
#'
#' @param x (`character(1)`)\cr
#'   One of `"latest"` or `"history"`. Default `"latest"`.
#' @returns A [data.table::data.table()] with the exchange rates.
#' @source <https://www.ecb.europa.eu/stats/policy_and_exchange_rates/euro_reference_exchange_rates/html/index.en.html>
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' ecb_fx_rates()
#' }
ecb_fx_rates <- function(x = "latest") {
  assert_choice(x, c("latest", "history"))
  url <- switch(
    x,
    latest = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref.zip",
    history = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.zip"
  )
  tf <- tempfile()
  on.exit(unlink(tf), add = TRUE)
  curl::curl_download(url, tf)
  dt <- fread(file = tf, sep = ",", na.strings = c("NA", "N/A"))
  fmt <- if (nrow(dt) > 1L) "%Y-%m-%d" else "%d %B %Y"
  Date <- NULL
  dt[, Date := as.Date(Date, fmt)]
  dt[, names(.SD) := map(.SD, as.numeric), .SDcols = !"Date"]
  dt <- dt |>
    melt(
      id.vars = "Date",
      variable.name = "currency",
      value.name = "rate",
      variable.factor = FALSE
    ) |>
    na.omit() |>
    setnames(tolower)
  dt[]
}

#' @rdname ecb_fx_rates
#' @export
ecb_euro_rates <- ecb_fx_rates

#' Fetch Bank of Canada foreign exchange rates
#'
#' Fetch the latest or historical foreign exchange reference rates from the Bank of Canada (BoC).
#'
#' @details
#' The recorded rates indicate the number of Canadian dollars required to buy a single unit of the
#' foreign currency. New rates are released by the Bank of Canada (BoC) daily at 4:30 pm. The Canada
#' Border Services (CBSA) retrieves these updates between 4:30 pm and 5 pm ET.
#'
#' BoC provides 23 foreign exchange rates. All other rates are maintained by the CBSA.
#'
#' Exchange rates from the BoC are updated daily in the system while other exchange rates are
#' updated by the CBSA at set intervals. The updated rates are available for retrieval between 7 pm
#' and 11:59 pm ET.
#'
#' As BoC publishes exchange rates every business day, it is recommended that exchange rate data be
#' retrieved on a daily basis. This retrieval should occur after 7 pm ET to ensure retrieval of the
#' latest updates.
#'
#' @param start_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   Start date of the data. Default `NULL`.
#' @param end_date (`NULL` | `Date(1)` | `character(1)`)\cr
#'   End date of the data. Default `NULL`.
#' @param limit (`NULL` | `integer(1)`)\cr
#'   Maximum number of records to return. Default `NULL` (all records).
#' @param skip (`NULL` | `integer(1)`)\cr
#'   Number of records to skip. Default `NULL` (do not skip any records).
#' @returns A [data.table::data.table()] with the exchange rates.
#' @source <https://www.cbsa-asfc.gc.ca/eservices/api/er-tc-api-eng.html>
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch latest exchange rates
#' boc_fx_rates()
#'
#' # fetch historical exchange rates
#' boc_fx_rates(start_date = "2021-10-22", end_date = "2021-10-23", limit = 10, skip = 2)
#' }
boc_fx_rates <- function(start_date = NULL, end_date = NULL, limit = NULL, skip = NULL) {
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)
  limit <- assert_count(limit, positive = TRUE, null.ok = TRUE, coerce = TRUE)
  skip <- assert_count(skip, null.ok = TRUE, coerce = TRUE)

  url <- "https://bcd-api-dca-ipa.cbsa-asfc.cloud-nuage.canada.ca/exchange-rate-lambda/exchange-rates" # nolint
  json <- request(url) |>
    req_user_agent(bbk_user_agent()) |>
    req_url_query(startDate = start_date, endDate = end_date, limit = limit, skip = skip) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_json()

  dt <- json$ForeignExchangeRates |>
    map(as.data.table) |>
    rbindlist() |>
    setnames(convert_camel_case)

  dt[, names(.SD) := map(.SD, \(x) unlist(x, use.names = FALSE)), .SDcols = is.list]
  rate <- NULL
  dt[, rate := as.numeric(rate)]
  dt[,
    names(.SD) := map(.SD, \(x) as.POSIXct(x, format = "%Y-%m-%dT%H:%M:%OSZ", tz = "UTC")),
    .SDcols = patterns("_timestamp$")
  ]
  dt[]
}

test_that("ecb_fx_rates works", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  expect_error(ecb_fx_rates("monthly"))
  expect_error(ecb_fx_rates(NA))
  expect_error(ecb_fx_rates(""))

  x <- ecb_fx_rates("latest")
  expect_data_table(x, min.rows = 1L, ncols = 3L)
  expect_named(x, c("date", "currency", "rate"))
  expect_date(x$date)
  expect_character(x$currency)
  expect_double(x$rate)

  x <- ecb_fx_rates("history")
  expect_data_table(x, min.rows = 31L)
  expect_named(x, c("date", "currency", "rate"))
  expect_date(x$date)
  expect_character(x$currency)
  expect_double(x$rate)
})

#' Fetch Banco de España (BdE) data
#'
#' Retrieve time series data from the BdE statistics API.
#'
#' @details
#' You can search for the series codes in the
#' [BIEST](https://app.bde.es/bie_www/bie_wwwias/xml/Arranque.html) application or in the tables
#' published by the Banco de España.
#'
#' @param key (`character()`)\cr
#'   The series keys to query.
#' @param time_range (`NULL` | `character(1)` | `integer(1)`)\cr
#'   The time range for the data. Can be an annual range (e.g., `2024`) or a frequency-based code:
#'   * Daily frequency (D): `"3M"` (last 3 months), `"12M"`, `"36M"`
#'   * Monthly frequency (M): `"30M"`, `"60M"`, `"MAX"` (entire series)
#'   * Quarterly frequency (Q): `"30M"`, `"60M"`, `"MAX"`
#'   * Annual frequency (A): `"60M"`, `"MAX"`
#'
#'   If `NULL` (default), returns the smallest range for the series frequency
#'   (e.g., `"30M"` for monthly series).
#' @param lang (`character(1)`)\cr
#'   Language to query, either `"en"` or `"es"`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.bde.es/webbe/en/estadisticas/recursos/api-estadisticas-bde.html>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bde_data("D_1NBAF472", time_range = "30M")
#' bde_data(c("DTNPDE2010_P0000P_PS_APU", "DTNSEC2010_S0000P_APU_SUMAMOVIL"), time_range = "MAX")
#' bde_data("DEEQ.N.ES.W1.S1.S1.T.B.G._Z._Z._Z.EUR._T._X.N.ALL", time_range = 2024)
#' }
bde_data <- function(key, time_range = NULL, lang = "en") {
  assert_character(key, min.chars = 1L)
  assert_period(time_range)
  assert_choice(lang, c("en", "es"))

  json <- bde(key, lang = lang, rango = time_range)
  parse_bde_data(json)
}

parse_bde_data <- function(json) {
  old_cols <- c(
    "serie",
    "descripcion",
    "descripcionCorta",
    "codFrecuencia",
    "decimales",
    "simbolo",
    "fechaInicio",
    "fechaFin",
    "fechas",
    "valores"
  )
  new_cols <- c(
    "key",
    "description",
    "title",
    "freq",
    "decimals",
    "symbol",
    "start_date",
    "end_date",
    "date",
    "value"
  )
  dt <- json |>
    map(function(x) {
      dt <- as.data.table(x[names(x) != "informacion"])
      meta <- x$informacion |>
        map(setDT) |>
        rbindlist() |>
        setnames(c("name", "value"))
      name <- NULL
      meta[, name := tolower(name)]
      meta[, name := chartr(" ", "_", tolower(name))]
      meta[, name := gsub("[()]", "", name)]
      exclude <- c(
        "name",
        "decimals",
        "first_value",
        "last_value",
        "number_of_observations",
        "min_value",
        "max_value"
      )
      meta <- meta[!name %in% exclude]
      meta <- transpose(meta, make.names = "name")
      setnames(meta, c("units", "description"), c("unit", "long_description"))
      dt[, names(meta) := meta]
    }) |>
    rbindlist(fill = TRUE) |>
    setnames(old_cols, new_cols)

  value <- NULL
  dt[, let(date = unlist(date, use.names = FALSE), value = unlist(value, use.names = FALSE))]
  dt[,
    names(.SD) := map(.SD, \(x) as.POSIXct(x, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")),
    .SDcols = patterns("date")
  ]
  setcolorder(dt, col_order, skip_absent = TRUE)
  dt[]
}

#' Fetch latest Banco de España (BdE) data
#'
#' Retrieve the most recently published value for one or more series from the BdE statistics API.
#'
#' @inheritParams bde_data
#' @returns A [data.table::data.table()] with the latest observation per series.
#' @source <https://www.bde.es/webbe/en/estadisticas/recursos/api-estadisticas-bde.html>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' bde_latest("D_1NBAF472")
#' bde_latest(c("D_1NBAF472", "DTNPDE2010_P0000P_PS_APU"))
#' }
bde_latest <- function(key, lang = "en") {
  assert_character(key, min.chars = 1L)
  assert_choice(lang, c("en", "es"))

  json <- bde(key, lang = lang, resource = "favoritas")

  old_cols <- c(
    "serie",
    "descripcionCorta",
    "codFrecuencia",
    "decimales",
    "simbolo",
    "tendencia",
    "fechaValor",
    "valor"
  )
  new_cols <- c("key", "title", "freq", "decimals", "symbol", "trend", "date", "value")
  dt <- json |>
    map(setDT) |>
    rbindlist() |>
    setnames(old_cols, new_cols)
  dt[, date := as.POSIXct(date, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")]
  setcolorder(dt, col_order, skip_absent = TRUE)
  dt[]
}

bde <- function(key, ..., lang, resource = "listaSeries") {
  url <- sprintf("https://app.bde.es/bierest/resources/srdatosapp/%s", resource)
  request(url) |>
    req_user_agent(bbk_user_agent()) |>
    req_url_query(idioma = lang, series = key, ..., .multi = "comma") |>
    req_error(body = bde_error_body) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform() |>
    resp_body_json()
}

bde_error_body <- function(resp) {
  content_type <- resp_content_type(resp)
  if (identical(content_type, "application/json")) {
    json <- resp_body_json(resp)
    msg <- c(json$errMsgUsr, json$errMsgDebug)
    docs <- "See docs at <https://www.bde.es/webbe/en/estadisticas/recursos/api-estadisticas-bde.html>" # nolint
    c(msg, docs)
  }
}

test_that("bde_data input validation works", {
  # series should be a character()
  expect_error(bde_data(1L))
  expect_error(bde_data(TRUE))
  expect_error(bde_data(NULL))
})

test_that("bde_latest input validation works", {
  expect_error(bde_latest(1L))
  expect_error(bde_latest(TRUE))
  expect_error(bde_latest(NULL))
  expect_error(bde_latest("abc", lang = "fr"))
})

test_that("bde_latest works", {
  json <- readRDS(test_path("fixtures", "bde-latest.rds"))
  dt <- rbindlist(map(json, setDT))
  # fmt: skip
  old_cols <- c("serie", "descripcionCorta", "codFrecuencia", "decimales", "simbolo", "tendencia", "fechaValor", "valor") # nolint
  new_cols <- c("key", "title", "freq", "decimals", "symbol", "trend", "date", "value")
  setnames(dt, old_cols, new_cols)
  dt[, date := as.POSIXct(date, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")]
  expect_data_table(dt, nrows = 2L)
  expect_names(names(dt), must.include = c("key", "title", "value", "date"))
  expect_posixct(dt$date)
})

test_that("parse_bde_data works", {
  json <- readRDS(test_path("fixtures", "bde-data.rds"))
  actual <- parse_bde_data(json)
  expect_data_table(actual, min.rows = 1L)
  expect_identical(
    unique(actual$key),
    c("DTNPDE2010_P0000P_PS_APU", "DTNSEC2010_S0000P_APU_SUMAMOVIL")
  )
  expect_posixct(actual$date)
  expect_false(any(unlist(map(actual, class)) == "list"))
})

#' Fetch Norges Bank (NoB) data
#'
#' Retrieve time series data from the Norges Bank SDMX Web Service.
#'
#' @param flow (`character(1)`)\cr
#'   The dataflow to query. See [nob_metadata()] for available dataflows.
#' @param key (`NULL` | `character(1)`)\cr
#'   The series key to query using dot-separated dimension values
#'   (e.g., `"B.USD.NOK.SP"`). Use `+` for multiple values in one dimension
#'   (e.g., `"B.USD+EUR.NOK.SP"`). If `NULL`, all data for the flow is returned.
#'   Default `NULL`.
#' @param start_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   Start date of the data (e.g., `"2024-01-01"` or `2024`). If `NULL`, no start date restriction
#'   is applied. Default `NULL`.
#' @param end_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   End date of the data, in the same format as start_period. If `NULL`, no end date restriction is
#'   applied. Default `NULL`.
#' @param first_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the start of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param last_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the end of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://www.norges-bank.no/en/topics/Statistics/open-data/>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch USD/NOK exchange rate
#' nob_data("EXR", "B.USD.NOK.SP", last_n = 5L)
#'
#' # fetch multiple exchange rates
#' nob_data("EXR", "B.USD+EUR+GBP.NOK.SP", start_period = "2024-01-01")
#'
#' # fetch policy rate
#' nob_data("IR", last_n = 5L)
#' }
nob_data <- function(
  flow,
  key = NULL,
  start_period = NULL,
  end_period = NULL,
  first_n = NULL,
  last_n = NULL
) {
  assert_string(flow, min.chars = 1L)
  assert_string(key, min.chars = 1L, null.ok = TRUE)
  assert_period(start_period)
  assert_period(end_period)
  first_n <- assert_count(first_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  last_n <- assert_count(last_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)

  resource <- sdmx_data_resource(flow, key)
  xml <- nob(
    resource,
    startPeriod = start_period,
    endPeriod = end_period,
    firstNObservations = first_n,
    lastNObservations = last_n
  )
  parse_nob_data(xml)
}

#' Fetch Norges Bank (NoB) metadata
#'
#' Retrieve metadata from the Norges Bank SDMX Web Service.
#'
#' @param type (`character(1)`)\cr
#'   The type of metadata to query.
#'   One of: `"datastructure"`, `"dataflow"`, `"codelist"`, or `"concept"`.
#' @param id (`NULL` | `character(1)`)\cr
#'   The id to query. Default `NULL`.
#' @param lang (`character(1)`)\cr
#'   Language for names, either `"en"` or `"no"`. Default `"en"`.
#' @returns A [data.table::data.table()] with the requested metadata.
#' @source <https://www.norges-bank.no/en/topics/Statistics/open-data/>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' nob_metadata("dataflow")
#' nob_metadata("datastructure")
#' nob_metadata("codelist", "CL_CURRENCY")
#' }
nob_metadata <- function(type, id = NULL, lang = "en") {
  assert_choice(type, c("datastructure", "dataflow", "codelist", "concept"))
  assert_string(id, min.chars = 1L, null.ok = TRUE)
  assert_choice(lang, c("en", "no"))

  xpath <- switch(
    type,
    datastructure = "//str:DataStructure",
    dataflow = "//str:Dataflow",
    codelist = "//str:Codelist",
    concept = "//str:ConceptScheme"
  )
  type <- if (type == "concept") "conceptscheme" else type
  resource <- if (is.null(id)) type else paste(type, "NB", toupper(id), sep = "/")
  xml <- nob(resource)
  entries <- xml2::xml_find_all(xml, xpath)
  sdmx_metadata(entries, lang)
}

#' Fetch Norges Bank (NoB) dimensions
#'
#' Retrieve the dimension structure for a given dataflow from the Norges Bank SDMX Web Service.
#'
#' @param id (`character(1)`)\cr
#'   The id of the data structure definition to query (e.g., `"NB_EXR"`).
#' @returns A [data.table::data.table()] with columns:
#'   \item{id}{The dimension id (e.g., `"FREQ"`, `"BASE_CUR"`)}
#'   \item{position}{The position of the dimension in the series key}
#'   \item{codelist}{The id of the associated codelist (e.g., `"CL_FREQ"`)}
#' @source <https://www.norges-bank.no/en/topics/Statistics/open-data/>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' nob_dimension("DSD_EXR")
#' }
nob_dimension <- function(id) {
  assert_string(id, min.chars = 1L)
  resource <- paste("datastructure", "NB", toupper(id), sep = "/")
  xml <- nob(resource)
  sdmx_dimension(xml)
}

parse_nob_data <- function(xml) {
  ns <- xml2::xml_ns(xml)
  series <- xml2::xml_find_all(xml, ".//Series", ns)
  res <- map(series, function(x) {
    attrs <- xml2::xml_attrs(x)
    nms <- tolower(names(attrs))
    names(attrs) <- nms

    obs <- xml2::xml_find_all(x, ".//Obs", ns)
    obs_attrs <- map(obs, xml2::xml_attrs)
    date <- map_chr(obs_attrs, "TIME_PERIOD")
    value <- as.numeric(map_chr(obs_attrs, "OBS_VALUE"))

    key <- paste(
      attrs[!nms %in% c("collection", "calculated", "decimals", "unit_mult")],
      collapse = "."
    )

    freq <- sdmx_freq(attrs[["freq"]])

    extra <- attrs[!nms %in% c("freq", "collection", "calculated", "decimals", "unit_mult")]
    data <- c(
      list(date = parse_date(date, freq), key = key, value = value, freq = freq),
      as.list(extra)
    )
    as.data.table(data)
  })
  res <- res |>
    rbindlist(fill = TRUE) |>
    setcolorder(col_order, skip_absent = TRUE)
  res[]
}

nob_error_body <- function(resp) {
  resp_body_string(resp, "UTF-8")
}

nob <- function(resource, ...) {
  sdmx_request("https://data.norges-bank.no/api", resource, nob_error_body, ...)
}

test_that("nob_data input validation works", {
  expect_error(nob_data(1L))
  expect_error(nob_data(TRUE))
  expect_error(nob_data(NULL))
  expect_error(nob_data(NA))
  expect_error(nob_data(c("EXR", "IR")))
  # key should be a character(1) or NULL
  expect_error(nob_data("EXR", 1L))
  expect_error(nob_data("EXR", TRUE))
  expect_error(nob_data("EXR", NA))
  # start_period
  expect_error(nob_data("EXR", "abc", start_period = TRUE))
  expect_error(nob_data("EXR", "abc", start_period = c("a", "b")))
  # end_period
  expect_error(nob_data("EXR", "abc", end_period = TRUE))
  expect_error(nob_data("EXR", "abc", end_period = c("a", "b")))
  # last_n
  expect_error(nob_data("EXR", "abc", last_n = "abc"))
  expect_error(nob_data("EXR", "abc", last_n = TRUE))
  expect_error(nob_data("EXR", "abc", last_n = -1L))
  expect_error(nob_data("EXR", "abc", last_n = 0L))
})

test_that("parse_nob_data works", {
  body <- xml2::read_xml(test_path("fixtures", "nob-data.xml"))
  actual <- parse_nob_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_date(actual$date)
  expect_numeric(actual$value)
  expect_true(all(c("date", "key", "value", "freq") %in% names(actual)))
})

test_that("nob_dimension input validation works", {
  expect_error(nob_dimension(1L))
  expect_error(nob_dimension(TRUE))
  expect_error(nob_dimension(NULL))
  expect_error(nob_dimension(c("a", "b")))
})

test_that("nob_metadata input validation works", {
  expect_error(nob_metadata(1L))
  expect_error(nob_metadata(TRUE))
  expect_error(nob_metadata("data"))
  expect_error(nob_metadata(c("datastructure", "dataflow")))
  expect_error(nob_metadata("dataflow", id = 1L))
  expect_error(nob_metadata("dataflow", id = TRUE))
  expect_error(nob_metadata("dataflow", lang = "fr"))
})

test_that("sdmx_metadata works for nob", {
  body <- xml2::read_xml(test_path("fixtures", "nob-metadata.xml"))
  entries <- xml2::xml_find_all(body, "//str:Dataflow")
  actual <- sdmx_metadata(entries)
  expect_data_table(actual, min.rows = 1L)
  expect_true(all(c("id", "name") %in% names(actual)))
  expect_true("EXR" %in% actual$id)
})

#' Fetch Banque de France (BdF) data
#'
#' Retrieve time series data from the BdF Webstat API.
#'
#' @param ... (`any`)\cr
#'   Extra arguments appended to the API request.
#'   Combined with the default arguments with [modifyList()].
#' @param key (`NULL` | `character(1)`)\cr
#'   The series key to query. Default `NULL`.
#' @param start_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   Start date of the data. Default `NULL`.
#' @param end_date (`NULL` | `character(1)` | `Date(1)`)\cr
#'   End date of the data. Default `NULL`.
#' @param lang (`character(1)`)\cr
#'   Language to query. Default `"en"`.
#' @param api_key (`character(1)`)\cr
#'   API key to use for the request. Defaults to the value returned by `bdf_key()`, which reads from
#'   the `BANQUEDEFRANCE_KEY` environment variable.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://webstat.banque-france.fr/en/pages/guide-migration-api/>
#' @family data
#' @export
#' @examples
#' \dontrun{
#' bdf_data(key = "CONJ2.M.R24.T.SM.0RG24.EFTPM100.10")
#'
#' # inflation rate
#' bdf_data(key = "ICP.M.FR.N.000000.4.ANR")
#'
#' # or with a date filter
#' bdf_data(key = "ICP.M.FR.N.000000.4.ANR", start_date = "2025-01-01", end_date = "2025-06-30")
#'
#' # advanced filter with where clause
#' bdf_data(key = "ICP.M.FR.N.000000.4.ANR", where = "time_period_start >= date'2025-01-01'")
#' }
bdf_data <- function(
  ...,
  key = NULL,
  start_date = NULL,
  end_date = NULL,
  lang = "en",
  api_key = bdf_key()
) {
  assert_string(key, min.chars = 1L, null.ok = TRUE)
  assert_string(lang, min.chars = 1L)
  assert_string(api_key, min.chars = 1L)
  start_date <- assert_dateish(start_date, null.ok = TRUE)
  end_date <- assert_dateish(end_date, null.ok = TRUE)

  if (!is.null(key)) {
    key <- sprintf('series_key:"%s"', key)
  }
  where <- character()
  if (!is.null(start_date)) {
    start_date <- sprintf("time_period_start >= date'%s'", start_date)
    where <- c(where, start_date)
  }
  if (!is.null(end_date)) {
    end_date <- sprintf("time_period_end <= date'%s'", end_date)
    where <- c(where, end_date)
  }
  where <- if (length(where) > 0L) paste(where, collapse = " and ") else NULL
  params <- list(refine = key, where = where, lang = lang)
  params <- utils::modifyList(params, list(...))

  dt <- do.call(bdf, c(list(resource = "observations/exports/csv", api_key = api_key), params))
  parse_bdf_data(dt)
}

#' Fetch Banque de France (BdF) datasets
#'
#' @inheritParams bdf_data
#' @inherit bdf_data return
#' @source <https://webstat.banque-france.fr/en/pages/guide-migration-api/>
#' @family data
#' @export
#' @examples
#' \dontrun{
#' bdf_dataset()
#'
#' # structure of a dataset
#' bdf_dataset(where = "dataset_id = 'CONJ2'")
#' }
bdf_dataset <- function(..., lang = "en") {
  assert_string(lang, min.chars = 1L)
  dt <- bdf(resource = "webstat-datasets/exports/csv", lang = lang, ...)
  parse_bdf_dataset(dt)
}

#' Fetch Banque de France (BdF) codelists
#'
#' @inheritParams bdf_data
#' @inherit bdf_data return
#' @source <https://webstat.banque-france.fr/en/pages/guide-migration-api/>
#' @family data
#' @export
#' @examples
#' \dontrun{
#' bdf_codelist()
#'
#' # filter for a specific codelist
#' bdf_codelist(where = "codelist_id = 'CL_FREQ'")
#' }
bdf_codelist <- function(..., lang = "en") {
  assert_string(lang, min.chars = 1L)
  bdf(resource = "codelists/exports/csv", lang = lang, ...)
}

parse_bdf_data <- function(dt) {
  cols <- names(dt)
  path_cols <- grepv("^path_", cols)
  if (length(path_cols) > 0L) {
    dt[, (path_cols) := map(mget(path_cols), \(x) strsplit(x, ",", fixed = TRUE))]
  }
  if ("observations_attributes_and_values" %in% cols) {
    observations_attributes_and_values <- NULL # nolint
    dt[,
      observations_attributes_and_values := gsub(
        '""',
        '"',
        observations_attributes_and_values,
        fixed = TRUE
      )
    ]
    dt[,
      observations_attributes_and_values := map(
        observations_attributes_and_values,
        jsonlite::fromJSON
      )
    ]
  }
  dt[]
}

parse_bdf_dataset <- function(dt) {
  cols <- names(dt)
  paths_cols <- grepv("^paths_", cols)
  if (length(paths_cols) > 0L) {
    dt[, (paths_cols) := map(mget(paths_cols), \(x) strsplit(x, ",", fixed = TRUE))]
  }
  codelist_cols <- grepv("_codelists$", cols)
  if (length(codelist_cols) > 0L) {
    dt[,
      (codelist_cols) := map(mget(codelist_cols), function(x) {
        map(gsub('""', '"', x, fixed = TRUE), jsonlite::fromJSON)
      })
    ]
  }
  dt[]
}

bdf <- function(resource, ..., api_key = bdf_key()) {
  tf <- tempfile()
  on.exit(unlink(tf), add = TRUE)
  request("https://webstat.banque-france.fr/api/explore/v2.1/catalog/datasets") |>
    req_url_path_append(resource) |>
    req_user_agent(bbk_user_agent()) |>
    req_headers(Authorization = paste("Apikey", api_key)) |>
    req_error(body = bdf_error_body) |>
    req_url_query(..., delimiter = ";", compressed = TRUE) |>
    req_bbk_retry() |>
    req_bbk_cache() |>
    req_perform(path = tf)
  fread(file = tf, sep = ";")
}

bdf_error_body <- function(resp) {
  content_type <- resp_content_type(resp)
  if (identical(content_type, "application/json")) {
    json <- resp_body_json(resp)
    msg <- c(json$error_key, json$message)
    docs <- "See docs at <https://webstat.banque-france.fr/en/pages/guide-migration-api>" # nolint
    c(msg, docs)
  }
}

bdf_key <- function() {
  key <- Sys.getenv("BANQUEDEFRANCE_KEY")
  if (nzchar(key)) {
    return(key)
  }
  if (is_testing()) {
    testthat::skip(sprintf("%s env var is not configured", "BDF"))
  }
  stop(
    "No API key found, please supply with `api_key` argument or with BANQUEDEFRANCE_KEY env var.",
    call. = FALSE
  )
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

#' Fetch European Central Bank (ECB) data
#'
#' Retrieve time series data from the ECB SDMX Web Service.
#'
#' @param flow (`character(1)`)\cr
#'   Flow to query.
#' @param key (`NULL` | `character()`)\cr
#'   The series keys to query.
#' @param start_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   Start date of the data. Supported formats:
#'   * YYYY for annual data (e.g., `2019`)
#'   * YYYY-S\[1-2\] for semi-annual data (e.g., `"2019-S1"`)
#'   * YYYY-Q\[1-4\] for quarterly data (e.g., `"2019-Q1"`)
#'   * YYYY-MM for monthly data (e.g., `"2019-01"`)
#'   * YYYY-W\[01-53\] for weekly data (e.g., `"2019-W01"`)
#'   * YYYY-MM-DD for daily and business data (e.g., `"2019-01-01"`)
#'
#'   If `NULL`, no start date restriction is applied (data retrieved from the earliest available
#'   date). Default `NULL`.
#' @param end_period (`NULL` | `character(1)` | `integer(1)`)\cr
#'   End date of the data, in the same format as start_period. If `NULL`, no end date restriction is
#'   applied (data retrieved up to the most recent available date). Default `NULL`.
#' @param first_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the start of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param last_n (`NULL` | `numeric(1)`)\cr
#'   Number of observations to retrieve from the end of the series. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @param updated_after (`NULL` | `character(1)` | `Date(1)` | `POSIXct(1)`)\cr
#'   Retrieve only observations updated after the given timestamp (e.g.,
#'   `"2024-06-01T00:00:00"`). Useful for incremental retrieval. If `NULL`, no restriction is
#'   applied. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested data.
#' @source <https://data.ecb.europa.eu/help/api/data>
#' @family data
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' # fetch US dollar/Euro exchange rate
#' ecb_data("EXR", "D.USD.EUR.SP00.A")
#'
#' # fetch data for multiple keys
#' ecb_data("EXR", c("D.USD", "JPY.EUR.SP00.A"))
#' }
ecb_data <- function(
  flow,
  key = NULL,
  start_period = NULL,
  end_period = NULL,
  first_n = NULL,
  last_n = NULL,
  updated_after = NULL
) {
  assert_string(flow, min.chars = 1L)
  assert_character(key, min.chars = 1L, null.ok = TRUE)
  assert_period(start_period)
  assert_period(end_period)
  first_n <- assert_count(first_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  last_n <- assert_count(last_n, null.ok = TRUE, positive = TRUE, coerce = TRUE)
  updated_after <- assert_timestampish(updated_after, null.ok = TRUE)

  resource <- sdmx_data_resource(flow, key, default_key = "all")
  xml <- ecb(
    resource = resource,
    startPeriod = start_period,
    endPeriod = end_period,
    firstNObservations = first_n,
    lastNObservations = last_n,
    updatedAfter = updated_after
  )
  parse_ecb_data(xml)
}

#' Fetch European Central Bank (ECB) metadata
#'
#' Retrieve metadata from the ECB time series database via the SDMX Web Service.
#'
#' @param type (`character(1)`)\cr
#'   The type of metadata to query. One of: `"datastructure"`, `"dataflow"`, `"codelist"`, or
#'   `"concept"`.
#' @param agency (`NULL` | `character(1)`)\cr
#'   The id of the agency to query. Default `NULL`.
#' @param id (`NULL` | `character(1)`)\cr
#'   The id of the resource to query. Default `NULL`.
#' @returns A [data.table::data.table()] with the requested metadata.
#' The columns are:
#'   \item{agency}{The agency of the metadata}
#'   \item{id}{The id of the metadata}
#'   \item{name}{The name of the metadata}
#' @source <https://data.ecb.europa.eu/help/api/metadata>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' ecb_metadata("datastructure")
#' ecb_metadata("datastructure", "ECB")
#' ecb_metadata("datastructure", "ECB", "ECB_EXR1")
#' ecb_metadata("datastructure", id = "ECB_EXR1")
#' }
ecb_metadata <- function(type, agency = NULL, id = NULL) {
  assert_choice(type, c("datastructure", "dataflow", "codelist", "concept"))
  args <- switch(
    type,
    datastructure = list("datastructure", "//str:DataStructure"),
    dataflow = list("dataflow", "//str:Dataflow"),
    codelist = list("codelist", "//str:Codelist"),
    concept = list("conceptscheme", "//str:ConceptScheme")
  )
  do.call(fetch_ecb_metadata, c(args, list(agency, id)))
}

#' Fetch European Central Bank (ECB) dimensions
#'
#' Retrieve the dimension structure for a given dataflow from the ECB SDMX Web Service.
#'
#' @param id (`character(1)`)\cr
#'   The id of the data structure definition to query (e.g., `"ECB_EXR1"`).
#' @returns A [data.table::data.table()] with columns:
#'   \item{id}{The dimension id (e.g., `"FREQ"`, `"CURRENCY"`)}
#'   \item{position}{The position of the dimension in the series key}
#'   \item{codelist}{The id of the associated codelist (e.g., `"CL_FREQ"`)}
#' @source <https://data.ecb.europa.eu/help/api/metadata>
#' @family metadata
#' @export
#' @examplesIf curl::has_internet()
#' \donttest{
#' ecb_dimension("ECB_EXR1")
#' }
ecb_dimension <- function(id) {
  assert_string(id, min.chars = 1L)
  resource <- paste("datastructure", "ECB", toupper(id), sep = "/")
  xml <- ecb(resource)
  sdmx_dimension(xml)
}

fetch_ecb_metadata <- function(resource, xpath, agency = NULL, id = NULL) {
  assert_string(agency, min.chars = 1L, null.ok = TRUE)
  assert_string(id, min.chars = 1L, null.ok = TRUE)

  agency <- if (!is.null(agency)) toupper(agency) else "all"
  id <- if (!is.null(id)) toupper(id) else "all"
  resource <- paste(resource, agency, id, sep = "/")
  xml <- ecb(resource)
  entries <- xml2::xml_find_all(xml, xpath)
  parse_ecb_metadata(entries)
}

parse_ecb_data <- function(xml) {
  series <- xml2::xml_find_all(xml, ".//generic:Series")
  res <- map(series, function(x) {
    series_key <- x |>
      xml2::xml_find_first(".//generic:SeriesKey") |>
      xml2::xml_children()
    nms <- series_key |>
      xml2::xml_attr("id") |>
      tolower()
    series_key <- series_key |>
      xml2::xml_attr("value") |>
      setNames(nms) |>
      as.list()

    attrs <- x |>
      xml2::xml_find_first(".//generic:Attributes") |>
      xml2::xml_children()
    nms <- attrs |>
      xml2::xml_attr("id") |>
      tolower()
    nms <- replace(nms, nms == "title_compl", "description")
    attrs <- attrs |>
      xml2::xml_attr("value") |>
      setNames(nms) |>
      as.list()

    data <- c(series_key, attrs)
    data$key <- paste(series_key, collapse = ".")
    data$freq <- sdmx_freq(data$freq)

    entries <- xml2::xml_find_all(x, ".//generic:Obs[generic:ObsValue]")
    data$date <- x |>
      xml2::xml_find_all(".//generic:ObsDimension") |>
      xml2::xml_attr("value") |>
      parse_date(data$freq)

    data$value <- entries |>
      xml2::xml_find_all(".//generic:ObsValue") |>
      xml2::xml_attr("value") |>
      as.numeric()

    as.data.table(data)
  })
  res <- res |>
    rbindlist(fill = TRUE) |>
    setcolorder(col_order, skip_absent = TRUE)
  res[]
}

parse_ecb_metadata <- function(entries) {
  agency <- NULL
  dt <- entries |>
    map(function(node) {
      dt <- sdmx_metadata(list(node))
      dt[, agency := xml2::xml_attr(node, "agencyID")]
    }) |>
    rbindlist() |>
    setcolorder(c("agency", "id", "name"))
  dt[]
}

ecb_error_body <- function(resp) {
  message <- resp_body_string(resp, "UTF-8")
  docs <- "See docs at <https://data.ecb.europa.eu/help/api/status-codes>"
  c(message, docs)
}

ecb <- function(resource, ...) {
  sdmx_request("https://data-api.ecb.europa.eu/service/", resource, ecb_error_body, ...)
}

test_that("ecb_data input validation works", {
  # flow should be a character(1)
  expect_error(ecb_data(1L, "D.USD.EUR.SP00.A"))
  expect_error(ecb_data(TRUE, "D.USD.EUR.SP00.A"))
  expect_error(ecb_data(NULL, "D.USD.EUR.SP00.A"))
  expect_error(ecb_data(NA, "D.USD.EUR.SP00.A"))
  expect_error(ecb_data(c("EXR", "EXR"), "D.USD.EUR.SP00.A"))
  # key should be a character() or NULL
  expect_error(ecb_data("EXR", 1L))
  expect_error(ecb_data("EXR", TRUE))
  expect_error(ecb_data("EXR", NA))
  expect_error(ecb_data("EXR", c("D.USD.EUR.SP00.A", "D.USD.EUR.SP00.A")))
  # start_period should be a character(1) or NULL
  expect_error(ecb_data("abcde", "abc", start_period = 1L))
  expect_error(ecb_data("abcde", "abc", start_period = TRUE))
  expect_error(ecb_data("abcde", "abc", start_period = c("a", "b")))
  # end_period should be a character(1) or NULL
  expect_error(ecb_data("abcde", "abc", TRUE))
  expect_error(ecb_data("abcde", "abc", end_period = 1L))
  expect_error(ecb_data("abcde", "abc", end_period = c("a", "b")))
  # first_n should be a whole number or NULL
  expect_error(ecb_data("abcde", "abc", first_n = "abc"))
  expect_error(ecb_data("abcde", "abc", first_n = 1:2))
  expect_error(ecb_data("abcde", "abc", first_n = TRUE))
  expect_error(ecb_data("abcde", "abc", first_n = 1.5))
  expect_error(ecb_data("abcde", "abc", first_n = -1L))
  expect_error(ecb_data("abcde", "abc", first_n = 0L))
  # last_n should be a whole number or NULL
  expect_error(ecb_data("abcde", "abc", last_n = "abc"))
  expect_error(ecb_data("abcde", "abc", last_n = TRUE))
  expect_error(ecb_data("abcde", "abc", last_n = 1:2))
  expect_error(ecb_data("abcde", "abc", last_n = 1.5))
  expect_error(ecb_data("abcde", "abc", last_n = -1L))
  expect_error(ecb_data("abcde", "abc", last_n = 0L))
  # updated_after should be NULL, Date, POSIXct, or ISO 8601 string
  expect_error(ecb_data("abcde", "abc", updated_after = 1L))
  expect_error(ecb_data("abcde", "abc", updated_after = TRUE))
  expect_error(ecb_data("abcde", "abc", updated_after = NA))
})

test_that("ecb_data passes updated_after as updatedAfter", {
  captured <- NULL
  httr2::local_mocked_responses(function(req) {
    captured <<- req
    httr2::response(200L, headers = "content-type: application/xml", body = charToRaw("<x/>"))
  })
  local_mocked_bindings(parse_ecb_data = function(xml) data.table())
  ecb_data("EXR", "D.USD.EUR.SP00.A", updated_after = as.Date("2024-06-01"))
  expect_match(captured$url, "updatedAfter=2024-06-01T00%3A00%3A00")
})

test_that("parse_bbk_data works", {
  body <- xml2::read_xml(test_path("fixtures", "ecb-data.xml"))
  actual <- parse_ecb_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_identical(unique(actual$key), "D.USD.EUR.SP00.A")
  expect_date(actual$date)

  body <- xml2::read_xml(test_path("fixtures", "ecb-data-fm.xml"))
  actual <- parse_ecb_data(body)
  expect_data_table(actual, min.rows = 1L)
  expect_gt(length(unique(actual$key)), 0L)
  expect_date(actual$date)
})

test_that("ecb_dimension input validation works", {
  expect_error(ecb_dimension(1L))
  expect_error(ecb_dimension(TRUE))
  expect_error(ecb_dimension(NULL))
  expect_error(ecb_dimension(c("a", "b")))
})

test_that("metadata input validation works", {
  # type should be one of datastructure, dataflow, codelist, concept
  expect_error(ecb_metadata(1L))
  expect_error(ecb_metadata(TRUE))
  expect_error(ecb_metadata(c("datastructure", "dataflow")))
  expect_error(ecb_metadata("data"))
  # id should be a character(1) or NULL
  expect_error(ecb_metadata("datastructure", id = 1L))
  expect_error(ecb_metadata("datastructure", id = TRUE))
  expect_error(ecb_metadata("datastructure", id = c("a", "b")))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/srb.R
\name{srb_calendar}
\alias{srb_calendar}
\title{Fetch Sveriges Riksbank (SRb) calendar days}
\source{
\url{https://developer.api.riksbank.se/}
}
\usage{
srb_calendar(start_date, end_date = NULL)
}
\arguments{
\item{start_date}{(\code{Date(1)} | \code{character(1)})\cr
Start date of the range (e.g., \code{"2024-01-01"}).}

\item{end_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
End date of the range. If \code{NULL}, data up to the latest available date is returned.
Default \code{NULL}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the calendar day information.
}
\description{
Retrieve Swedish banking calendar information from the Sveriges Riksbank SWEA API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
srb_calendar("2024-01-01", "2024-01-31")
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other metadata:
\code{\link[=bbk_dimension]{bbk_dimension()}},
\code{\link[=bbk_metadata]{bbk_metadata()}},
\code{\link[=bdp_dataset]{bdp_dataset()}},
\code{\link[=bdp_dimension]{bdp_dimension()}},
\code{\link[=bdp_domain]{bdp_domain()}},
\code{\link[=bdp_series]{bdp_series()}},
\code{\link[=bis_dimension]{bis_dimension()}},
\code{\link[=bis_metadata]{bis_metadata()}},
\code{\link[=boj_metadata]{boj_metadata()}},
\code{\link[=ecb_dimension]{ecb_dimension()}},
\code{\link[=ecb_metadata]{ecb_metadata()}},
\code{\link[=nob_dimension]{nob_dimension()}},
\code{\link[=nob_metadata]{nob_metadata()}},
\code{\link[=onb_dimension]{onb_dimension()}},
\code{\link[=onb_frequency]{onb_frequency()}},
\code{\link[=onb_hierarchy]{onb_hierarchy()}},
\code{\link[=onb_metadata]{onb_metadata()}},
\code{\link[=onb_toc]{onb_toc()}},
\code{\link[=snb_dimension]{snb_dimension()}},
\code{\link[=srb_series]{srb_series()}}
}
\concept{metadata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bde.R
\name{bde_data}
\alias{bde_data}
\title{Fetch Banco de España (BdE) data}
\source{
\url{https://www.bde.es/webbe/en/estadisticas/recursos/api-estadisticas-bde.html}
}
\usage{
bde_data(key, time_range = NULL, lang = "en")
}
\arguments{
\item{key}{(\code{character()})\cr
The series keys to query.}

\item{time_range}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
The time range for the data. Can be an annual range (e.g., \code{2024}) or a frequency-based code:
\itemize{
\item Daily frequency (D): \code{"3M"} (last 3 months), \code{"12M"}, \code{"36M"}
\item Monthly frequency (M): \code{"30M"}, \code{"60M"}, \code{"MAX"} (entire series)
\item Quarterly frequency (Q): \code{"30M"}, \code{"60M"}, \code{"MAX"}
\item Annual frequency (A): \code{"60M"}, \code{"MAX"}
}

If \code{NULL} (default), returns the smallest range for the series frequency
(e.g., \code{"30M"} for monthly series).}

\item{lang}{(\code{character(1)})\cr
Language to query, either \code{"en"} or \code{"es"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the BdE statistics API.
}
\details{
You can search for the series codes in the
\href{https://app.bde.es/bie_www/bie_wwwias/xml/Arranque.html}{BIEST} application or in the tables
published by the Banco de España.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
bde_data("D_1NBAF472", time_range = "30M")
bde_data(c("DTNPDE2010_P0000P_PS_APU", "DTNSEC2010_S0000P_APU_SUMAMOVIL"), time_range = "MAX")
bde_data("DEEQ.N.ES.W1.S1.S1.T.B.G._Z._Z._Z.EUR._T._X.N.ALL", time_range = 2024)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boc.R
\name{boc_metadata}
\alias{boc_metadata}
\title{Fetch Bank of Canada (BoC) metadata (details)}
\source{
\url{https://www.bankofcanada.ca/valet/docs}
}
\usage{
boc_metadata(group_name = NULL, series_name = NULL)
}
\arguments{
\item{group_name}{(\code{NULL} | \code{character(1)})\cr
Name of the group. Only one of \code{group_name} or \code{series_name} can be used.}

\item{series_name}{(\code{NULL} | \code{character()})\cr
Name of the series.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Fetch Bank of Canada (BoC) metadata (details)
}
\examples{
\dontrun{
boc_metadata(group_name = "FX_RATES_DAILY")
boc_metadata(series_name = "FXUSDCAD")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache}
\alias{cache}
\alias{bbk_cache_dir}
\alias{bbk_cache_clear}
\title{Get or manage the bbk API cache}
\usage{
bbk_cache_dir()

bbk_cache_clear()
}
\description{
\code{bbk_cache_dir()} returns the path where cached API responses are stored.
\code{bbk_cache_clear()} clears all cached responses.
}
\details{
The cache is only used when enabled with \code{options(bbk.cache = TRUE)}.
Cached responses are stored for 1 day by default, but this can be customized with
\code{options(bbk.cache_max_age = seconds)}.
}
\examples{
\dontrun{
# enable caching
options(bbk.cache = TRUE)

# view cache location
bbk_cache_dir()

# clear the cache
bbk_cache_clear()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bbk-package.R
\docType{package}
\name{bbk-package}
\alias{bbk}
\alias{bbk-package}
\title{bbk: Client for Central Bank APIs}
\description{
A client for retrieving data and metadata from central bank APIs including 'Banco de España' (BdE), 'Banco de Portugal' (BdP), 'Bank for International Settlements' (BIS), 'Bank of Canada' (BoC), 'Bank of England' (BoE), 'Bank of Japan' (BoJ), 'Banque de France' (BdF), 'Deutsche Bundesbank' (BBk), 'European Central Bank' (ECB), 'National Bank of Poland' (NBP), 'Norges Bank' (NoB), 'Oesterreichische Nationalbank' (OeNB), 'Sveriges Riksbank' (SRb), and 'Swiss National Bank' (SNB).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://m-muecke.github.io/bbk/}
  \item \url{https://github.com/m-muecke/bbk}
  \item Report bugs at \url{https://github.com/m-muecke/bbk/issues}
}

}
\author{
\strong{Maintainer}: Maximilian Mücke \email{muecke.maximilian@gmail.com} (\href{https://orcid.org/0009-0000-9432-9795}{ORCID})

Authors:
\itemize{
  \item Maximilian Mücke \email{muecke.maximilian@gmail.com} (\href{https://orcid.org/0009-0000-9432-9795}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snb.R
\name{snb_data}
\alias{snb_data}
\title{Fetch Swiss National Bank (SNB) data}
\source{
\url{https://data.snb.ch/en}
}
\usage{
snb_data(key, start_date = NULL, end_date = NULL, lang = "en")
}
\arguments{
\item{key}{(\code{character(1)})\cr
The series key to query.}

\item{start_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
Start date of the data.}

\item{end_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
End date of the data.}

\item{lang}{(\code{character(1)})\cr
Language to query, either \code{"en"} or \code{"de"}. Default \code{"en"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the SNB data portal.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
snb_data("rendopar")

# or filter for date range
snb_data("rendopar", "2020-01-01", "2020-12-31")
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bdp.R
\name{bdp_data}
\alias{bdp_data}
\title{Fetch Banco de Portugal (BdP) data}
\source{
\url{https://bpstat.bportugal.pt/data/docs}
}
\usage{
bdp_data(
  domain_id,
  dataset_id,
  series_ids = NULL,
  start_date = NULL,
  end_date = NULL,
  last_n = NULL,
  updated_after = NULL,
  lang = "en"
)
}
\arguments{
\item{domain_id}{(\code{integer(1)})\cr
The domain ID. Use \code{\link[=bdp_domain]{bdp_domain()}} to list available domains.}

\item{dataset_id}{(\code{character(1)})\cr
The dataset ID within the domain.}

\item{series_ids}{(\code{NULL} | \code{integer()})\cr
Optional series IDs to filter the dataset.}

\item{start_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
Start date of the data.}

\item{end_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
End date of the data.}

\item{last_n}{(\code{NULL} | \code{integer(1)})\cr
Return only the last \code{n} observations per series.}

\item{updated_after}{(\code{NULL} | \code{character(1)} | \code{Date(1)} | \code{POSIXct(1)})\cr
Retrieve only observations published after the given timestamp (e.g.,
\code{"2024-06-01T00:00:00"}). Useful for incremental retrieval. If \code{NULL}, no restriction is
applied. Default \code{NULL}.}

\item{lang}{(\code{character(1)})\cr
Language for labels, either \code{"en"} or \code{"pt"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the BPstat API.
}
\details{
The BPstat API uses a two-step workflow: first look up the series metadata with
\code{\link[=bdp_series]{bdp_series()}} to find the \code{domain_id} and \code{dataset_id}, then use those to fetch the actual
observations.

You can browse available series at the \href{https://bpstat.bportugal.pt}{BPstat portal}.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# Portuguese GDP (annual, current prices)
bdp_data(54L, "ce3e4e50cda325537eff729ef64037cd", series_ids = 12518356L)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boc.R
\name{boc_catalog}
\alias{boc_catalog}
\title{Fetch Bank of Canada (BoC) available series or group}
\source{
\url{https://www.bankofcanada.ca/valet/docs}
}
\usage{
boc_catalog(type = "groups")
}
\arguments{
\item{type}{(\code{character(1)})\cr
Set of data to return. One of \code{"groups"} or \code{"series"}. Default \code{"groups"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Access all available series or groups from the Bank of Canada Valet API.
}
\examples{
\dontrun{
catalog <- boc_catalog()
head(catalog)

# filter for effective exchange rate series
dt <- catalog[grepl("CEER", label)]
head(dt)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fx-rates.R
\name{boc_fx_rates}
\alias{boc_fx_rates}
\title{Fetch Bank of Canada foreign exchange rates}
\source{
\url{https://www.cbsa-asfc.gc.ca/eservices/api/er-tc-api-eng.html}
}
\usage{
boc_fx_rates(start_date = NULL, end_date = NULL, limit = NULL, skip = NULL)
}
\arguments{
\item{start_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
Start date of the data. Default \code{NULL}.}

\item{end_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
End date of the data. Default \code{NULL}.}

\item{limit}{(\code{NULL} | \code{integer(1)})\cr
Maximum number of records to return. Default \code{NULL} (all records).}

\item{skip}{(\code{NULL} | \code{integer(1)})\cr
Number of records to skip. Default \code{NULL} (do not skip any records).}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the exchange rates.
}
\description{
Fetch the latest or historical foreign exchange reference rates from the Bank of Canada (BoC).
}
\details{
The recorded rates indicate the number of Canadian dollars required to buy a single unit of the
foreign currency. New rates are released by the Bank of Canada (BoC) daily at 4:30 pm. The Canada
Border Services (CBSA) retrieves these updates between 4:30 pm and 5 pm ET.

BoC provides 23 foreign exchange rates. All other rates are maintained by the CBSA.

Exchange rates from the BoC are updated daily in the system while other exchange rates are
updated by the CBSA at set intervals. The updated rates are available for retrieval between 7 pm
and 11:59 pm ET.

As BoC publishes exchange rates every business day, it is recommended that exchange rate data be
retrieved on a daily basis. This retrieval should occur after 7 pm ET to ensure retrieval of the
latest updates.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# fetch latest exchange rates
boc_fx_rates()

# fetch historical exchange rates
boc_fx_rates(start_date = "2021-10-22", end_date = "2021-10-23", limit = 10, skip = 2)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bbk.R
\name{bbk_dimension}
\alias{bbk_dimension}
\title{Fetch Deutsche Bundesbank (BBk) dimensions}
\source{
\url{https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-metadata}
}
\usage{
bbk_dimension(id)
}
\arguments{
\item{id}{(\code{character(1)})\cr
The id of the data structure definition to query (e.g., \code{"BBK_BBSIS"}).}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with columns:
\item{id}{The dimension id (e.g., \code{"BBK_STD_FREQ"}, \code{"BBK_STD_AREA"})}
\item{position}{The position of the dimension in the series key}
\item{codelist}{The id of the associated codelist}
}
\description{
Retrieve the dimension structure for a given dataflow from the Bundesbank SDMX Web Service.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
bbk_dimension("BBK_ERX")
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other metadata:
\code{\link[=bbk_metadata]{bbk_metadata()}},
\code{\link[=bdp_dataset]{bdp_dataset()}},
\code{\link[=bdp_dimension]{bdp_dimension()}},
\code{\link[=bdp_domain]{bdp_domain()}},
\code{\link[=bdp_series]{bdp_series()}},
\code{\link[=bis_dimension]{bis_dimension()}},
\code{\link[=bis_metadata]{bis_metadata()}},
\code{\link[=boj_metadata]{boj_metadata()}},
\code{\link[=ecb_dimension]{ecb_dimension()}},
\code{\link[=ecb_metadata]{ecb_metadata()}},
\code{\link[=nob_dimension]{nob_dimension()}},
\code{\link[=nob_metadata]{nob_metadata()}},
\code{\link[=onb_dimension]{onb_dimension()}},
\code{\link[=onb_frequency]{onb_frequency()}},
\code{\link[=onb_hierarchy]{onb_hierarchy()}},
\code{\link[=onb_metadata]{onb_metadata()}},
\code{\link[=onb_toc]{onb_toc()}},
\code{\link[=snb_dimension]{snb_dimension()}},
\code{\link[=srb_calendar]{srb_calendar()}},
\code{\link[=srb_series]{srb_series()}}
}
\concept{metadata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boc.R
\name{boc_data}
\alias{boc_data}
\title{Fetch Bank of Canada (BoC) data}
\source{
\url{https://www.bankofcanada.ca/valet/docs}
}
\usage{
boc_data(
  group_name = NULL,
  series_name = NULL,
  start_date = NULL,
  end_date = NULL
)
}
\arguments{
\item{group_name}{(\code{NULL} | \code{character(1)})\cr
Name of the group. Only one of \code{group_name} or \code{series_name} can be used.}

\item{series_name}{(\code{NULL} | \code{character()})\cr
Name of the series.}

\item{start_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
Start date of the data. Default \code{NULL}.}

\item{end_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
End date of the data. Default \code{NULL}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the Bank of Canada Valet API.
}
\examples{
\dontrun{
# fetch all data for a single group
dt <- boc_data(group_name = "FX_RATES_DAILY")
head(dt)

# or for multiple series ids
dt <- boc_data(
  series_name = c("FXUSDCAD", "FXEURCAD"),
  start_date = "2023-01-23",
  end_date = "2023-07-19"
)
head(dt)
}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fx-rates.R
\name{ecb_fx_rates}
\alias{ecb_fx_rates}
\alias{ecb_euro_rates}
\title{Fetch Euro foreign exchange reference rates}
\source{
\url{https://www.ecb.europa.eu/stats/policy_and_exchange_rates/euro_reference_exchange_rates/html/index.en.html}
}
\usage{
ecb_fx_rates(x = "latest")

ecb_euro_rates(x = "latest")
}
\arguments{
\item{x}{(\code{character(1)})\cr
One of \code{"latest"} or \code{"history"}. Default \code{"latest"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the exchange rates.
}
\description{
Fetch the latest or historical Euro foreign exchange reference rates from the
European Central Bank (ECB).
}
\details{
Note you can achieve the same by calling the \code{\link[=ecb_data]{ecb_data()}} function with the right
parameters for each currency.

The reference rates are usually updated at around 16:00 CET every working day, except on \href{https://www.ecb.europa.eu/ecb/contacts/working-hours/html/index.en.html}{TARGET closing days}.

They are based on the daily concertation procedure between central banks across
Europe, which normally takes place around 14:10 CET. The reference rates are
published for information purposes only.
Using the rates for transaction purposes is strongly discouraged.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
ecb_fx_rates()
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nob.R
\name{nob_data}
\alias{nob_data}
\title{Fetch Norges Bank (NoB) data}
\source{
\url{https://www.norges-bank.no/en/topics/Statistics/open-data/}
}
\usage{
nob_data(
  flow,
  key = NULL,
  start_period = NULL,
  end_period = NULL,
  first_n = NULL,
  last_n = NULL
)
}
\arguments{
\item{flow}{(\code{character(1)})\cr
The dataflow to query. See \code{\link[=nob_metadata]{nob_metadata()}} for available dataflows.}

\item{key}{(\code{NULL} | \code{character(1)})\cr
The series key to query using dot-separated dimension values
(e.g., \code{"B.USD.NOK.SP"}). Use \code{+} for multiple values in one dimension
(e.g., \code{"B.USD+EUR.NOK.SP"}). If \code{NULL}, all data for the flow is returned.
Default \code{NULL}.}

\item{start_period}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
Start date of the data (e.g., \code{"2024-01-01"} or \code{2024}). If \code{NULL}, no start date restriction
is applied. Default \code{NULL}.}

\item{end_period}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
End date of the data, in the same format as start_period. If \code{NULL}, no end date restriction is
applied. Default \code{NULL}.}

\item{first_n}{(\code{NULL} | \code{numeric(1)})\cr
Number of observations to retrieve from the start of the series. If \code{NULL}, no restriction is
applied. Default \code{NULL}.}

\item{last_n}{(\code{NULL} | \code{numeric(1)})\cr
Number of observations to retrieve from the end of the series. If \code{NULL}, no restriction is
applied. Default \code{NULL}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the Norges Bank SDMX Web Service.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# fetch USD/NOK exchange rate
nob_data("EXR", "B.USD.NOK.SP", last_n = 5L)

# fetch multiple exchange rates
nob_data("EXR", "B.USD+EUR+GBP.NOK.SP", start_period = "2024-01-01")

# fetch policy rate
nob_data("IR", last_n = 5L)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bbk.R
\name{bbk_data}
\alias{bbk_data}
\title{Fetch Deutsche Bundesbank (BBk) data}
\source{
\url{https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-data}
}
\usage{
bbk_data(
  flow,
  key = NULL,
  start_period = NULL,
  end_period = NULL,
  first_n = NULL,
  last_n = NULL,
  updated_after = NULL
)
}
\arguments{
\item{flow}{(\code{character(1)})\cr
The flow to query, 5-8 characters. See \code{\link[=bbk_metadata]{bbk_metadata()}} for available dataflows.}

\item{key}{(\code{NULL} | \code{character()})\cr
The series keys to query.}

\item{start_period}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
The start date of the data. Supported formats:
\itemize{
\item YYYY for annual data (e.g., `2019``)
\item YYYY-S[1-2] for semi-annual data (e.g., \code{"2019-S1"})
\item YYYY-Q[1-4] for quarterly data (e.g., \code{"2019-Q1"})
\item YYYY-MM for monthly data (e.g., \code{"2019-01"})
\item YYYY-W[01-53] for weekly data (e.g., \code{"2019-W01"})
\item YYYY-MM-DD for daily and business data (e.g., \code{"2019-01-01"})
}

If \code{NULL}, no start date restriction is applied (data retrieved from the earliest available
date). Default \code{NULL}.}

\item{end_period}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
The end date of the data, in the same format as start_period. If \code{NULL}, no end date
restriction is applied (data retrieved up to the most recent available date). Default \code{NULL}.}

\item{first_n}{(\code{NULL} | \code{numeric(1)})\cr
Number of observations to retrieve from the start of the series. If \code{NULL}, no restriction is
applied. Default \code{NULL}.}

\item{last_n}{(\code{NULL} | \code{numeric(1)})\cr
Number of observations to retrieve from the end of the series. If \code{NULL}, no restriction is
applied. Default \code{NULL}.}

\item{updated_after}{(\code{NULL} | \code{character(1)} | \code{Date(1)} | \code{POSIXct(1)})\cr
Retrieve only observations updated after the given timestamp (e.g.,
\code{"2024-06-01T00:00:00"}). Useful for incremental retrieval. If \code{NULL}, no restriction is
applied. Default \code{NULL}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the Bundesbank SDMX Web Service.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# fetch all data for a given flow and key
data <- bbk_data("BBSIS", "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A")
head(data)

# fetch data for multiple keys
data <- bbk_data("BBEX3", c("M.ISK.EUR", "USD.CA.AC.A01"))
head(data)

# specified period (start date-end date) for daily data
data <- bbk_data(
  "BBSIS", "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A",
  start_period = "2020-01-01",
  end_period = "2020-08-01"
)
head(data)

# or only specify the start date
data <- bbk_data(
  "BBSIS", "D.I.ZAR.ZI.EUR.S1311.B.A604.R10XX.R.A.A._Z._Z.A",
  start_period = "2024-04-01"
)
head(data)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onb.R
\name{onb_hierarchy}
\alias{onb_hierarchy}
\title{Fetch Österreichische Nationalbank (OeNB) hierarchy}
\usage{
onb_hierarchy(hier_id, lang = "en")
}
\arguments{
\item{hier_id}{(\code{integer(1)})\cr
Hierarchy id to query.}

\item{lang}{(\code{character(1)})\cr
Language to query. Default \code{"en"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Fetch Österreichische Nationalbank (OeNB) hierarchy
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
onb_hierarchy(hier_id = 11)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other metadata:
\code{\link[=bbk_dimension]{bbk_dimension()}},
\code{\link[=bbk_metadata]{bbk_metadata()}},
\code{\link[=bdp_dataset]{bdp_dataset()}},
\code{\link[=bdp_dimension]{bdp_dimension()}},
\code{\link[=bdp_domain]{bdp_domain()}},
\code{\link[=bdp_series]{bdp_series()}},
\code{\link[=bis_dimension]{bis_dimension()}},
\code{\link[=bis_metadata]{bis_metadata()}},
\code{\link[=boj_metadata]{boj_metadata()}},
\code{\link[=ecb_dimension]{ecb_dimension()}},
\code{\link[=ecb_metadata]{ecb_metadata()}},
\code{\link[=nob_dimension]{nob_dimension()}},
\code{\link[=nob_metadata]{nob_metadata()}},
\code{\link[=onb_dimension]{onb_dimension()}},
\code{\link[=onb_frequency]{onb_frequency()}},
\code{\link[=onb_metadata]{onb_metadata()}},
\code{\link[=onb_toc]{onb_toc()}},
\code{\link[=snb_dimension]{snb_dimension()}},
\code{\link[=srb_calendar]{srb_calendar()}},
\code{\link[=srb_series]{srb_series()}}
}
\concept{metadata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bdf.R
\name{bdf_data}
\alias{bdf_data}
\title{Fetch Banque de France (BdF) data}
\source{
\url{https://webstat.banque-france.fr/en/pages/guide-migration-api/}
}
\usage{
bdf_data(
  ...,
  key = NULL,
  start_date = NULL,
  end_date = NULL,
  lang = "en",
  api_key = bdf_key()
)
}
\arguments{
\item{...}{(\code{any})\cr
Extra arguments appended to the API request.
Combined with the default arguments with \code{\link[=modifyList]{modifyList()}}.}

\item{key}{(\code{NULL} | \code{character(1)})\cr
The series key to query. Default \code{NULL}.}

\item{start_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
Start date of the data. Default \code{NULL}.}

\item{end_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
End date of the data. Default \code{NULL}.}

\item{lang}{(\code{character(1)})\cr
Language to query. Default \code{"en"}.}

\item{api_key}{(\code{character(1)})\cr
API key to use for the request. Defaults to the value returned by \code{bdf_key()}, which reads from
the \code{BANQUEDEFRANCE_KEY} environment variable.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the BdF Webstat API.
}
\examples{
\dontrun{
bdf_data(key = "CONJ2.M.R24.T.SM.0RG24.EFTPM100.10")

# inflation rate
bdf_data(key = "ICP.M.FR.N.000000.4.ANR")

# or with a date filter
bdf_data(key = "ICP.M.FR.N.000000.4.ANR", start_date = "2025-01-01", end_date = "2025-06-30")

# advanced filter with where clause
bdf_data(key = "ICP.M.FR.N.000000.4.ANR", where = "time_period_start >= date'2025-01-01'")
}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/srb.R
\name{srb_cross_rates}
\alias{srb_cross_rates}
\title{Fetch Sveriges Riksbank (SRb) cross rates}
\source{
\url{https://developer.api.riksbank.se/}
}
\usage{
srb_cross_rates(series1, series2, start_date, end_date = NULL)
}
\arguments{
\item{series1}{(\code{character(1)})\cr
The first series ID (e.g., \code{"SEKUSDPMI"}).}

\item{series2}{(\code{character(1)})\cr
The second series ID (e.g., \code{"SEKEURPMI"}).}

\item{start_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
Start date of the data (e.g., \code{"2024-01-01"}). If \code{NULL}, all available data is returned.
Default \code{NULL}.}

\item{end_date}{(\code{NULL} | \code{Date(1)} | \code{character(1)})\cr
End date of the data, in the same format as start_date. If \code{NULL}, data up to the latest
available date is returned. Default \code{NULL}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the cross rate data.
}
\description{
Compute cross exchange rates between two currency series from the Sveriges Riksbank SWEA API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# USD/EUR cross rate
srb_cross_rates("SEKUSDPMI", "SEKEURPMI", start_date = "2024-01-01", end_date = "2024-01-31")
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boj.R
\name{boj_data}
\alias{boj_data}
\title{Fetch Bank of Japan (BoJ) data}
\source{
\url{https://www.stat-search.boj.or.jp/index_en.html}
}
\usage{
boj_data(db, code, start_date = NULL, end_date = NULL, lang = "en")
}
\arguments{
\item{db}{(\code{character(1)})\cr
The database code to query (e.g., \code{"FM08"} for foreign exchange rates). See the
\href{https://www.stat-search.boj.or.jp/info/api_manual_en.pdf}{API manual} for available databases.}

\item{code}{(\code{character()})\cr
One or more series codes to query (e.g., \code{"FXERD01"} for USD/JPY spot rate). Maximum 250
codes per request. All codes must have the same frequency. Use \code{\link[=boj_metadata]{boj_metadata()}} to find
available codes.}

\item{start_date}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
Start date of the data. Format depends on frequency: \code{"YYYYMMDD"} or \code{YYYY} for daily,
\code{"YYYYMM"} for monthly, \code{"YYYYQQ"} for quarterly (where QQ is 01-04), \code{"YYYY"} for annual.
If \code{NULL}, all available data is returned. Default \code{NULL}.}

\item{end_date}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
End date of the data, in the same format as start_date. If \code{NULL}, data up to the latest
available date is returned. Default \code{NULL}.}

\item{lang}{(\code{character(1)})\cr
Language for series names, either \code{"en"} or \code{"jp"}. Default \code{"en"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the Bank of Japan Statistics API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# fetch USD/JPY exchange rate
boj_data("FM08", "FXERD01", start_date = "202401")

# fetch multiple exchange rates
boj_data("FM08", c("FXERD01", "FXERD02"), start_date = "202401")
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onb.R
\name{onb_data}
\alias{onb_data}
\title{Fetch Österreichische Nationalbank (OeNB) data}
\source{
\url{https://www.oenb.at/en/Statistics/User-Defined-Tables/webservice.html}
}
\usage{
onb_data(
  hier_id,
  key,
  ...,
  start_period = NULL,
  end_period = NULL,
  freq = NULL,
  lang = "en"
)
}
\arguments{
\item{hier_id}{(\code{integer(1)})\cr
Hierarchy id to query.}

\item{key}{(\code{character()})\cr
The series keys to query.}

\item{...}{(\code{any})\cr
Additional parameters to pass to the API.}

\item{start_period}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
Start date of the data.}

\item{end_period}{(\code{NULL} | \code{character(1)} | \code{integer(1)})\cr
End date of the data.}

\item{freq}{(\code{NULL} | \code{character(1)})\cr
Frequency of the data.}

\item{lang}{(\code{character(1)})\cr
Language to query. Default \code{"en"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve time series data from the OeNB Web Service.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
onb_data(hier_id = 11, key = "VDBFKBSC217000")

# Loans to euro area residents, since 2000:
onb_data(hier_id = 11, key = "VDBFKBSC217000", start_period = "2000-01-01")

# Austrian imports and exports of goods from/to Germany, 2002–2012, annual frequency:
onb_data(hier_id = 901, key = "VDBQZA1000", start_period = 2002, end_period = 2012, freq = "A")

# Number of Austrian banks' subsidiaries abroad and in the EU, from 2005, semiannual:
onb_data(
  hier_id = 321,
  key = c("VDBKISDANZTAU", "VDBKISDANZTEU"),
  start_period = 200501,
  freq = "H"
)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bbk.R
\name{bbk_series}
\alias{bbk_series}
\title{Fetch the Deutsche Bundesbank (BBk) series}
\source{
\url{https://www.bundesbank.de/en/statistics/time-series-databases/help-for-sdmx-web-service/web-service-interface-data}
}
\usage{
bbk_series(key)
}
\arguments{
\item{key}{(\code{NULL} | \code{character()})\cr
The series keys to query.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Retrieve a single series by its key via the Bundesbank SDMX Web Service.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
bbk_series("BBEX3.M.DKK.EUR.BB.AC.A01")
bbk_series("BBAF3.Q.F41.S121.DE.S1.W0.LE.N._X.B")
bbk_series("BBBK11.D.TTA000")
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=bbk_data]{bbk_data()}} for an endpoint with more options.

Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bdf.R
\name{bdf_codelist}
\alias{bdf_codelist}
\title{Fetch Banque de France (BdF) codelists}
\source{
\url{https://webstat.banque-france.fr/en/pages/guide-migration-api/}
}
\usage{
bdf_codelist(..., lang = "en")
}
\arguments{
\item{...}{(\code{any})\cr
Extra arguments appended to the API request.
Combined with the default arguments with \code{\link[=modifyList]{modifyList()}}.}

\item{lang}{(\code{character(1)})\cr
Language to query. Default \code{"en"}.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with the requested data.
}
\description{
Fetch Banque de France (BdF) codelists
}
\examples{
\dontrun{
bdf_codelist()

# filter for a specific codelist
bdf_codelist(where = "codelist_id = 'CL_FREQ'")
}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_fx_rates]{nbp_fx_rates()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nbp.R
\name{nbp_fx_rates}
\alias{nbp_fx_rates}
\title{Fetch National Bank of Poland (NBP) exchange rates}
\source{
\url{https://api.nbp.pl/en.html}
}
\usage{
nbp_fx_rates(
  table,
  code = NULL,
  start_date = NULL,
  end_date = NULL,
  last_n = NULL
)
}
\arguments{
\item{table}{(\code{character(1)})\cr
Table type: \code{"a"} (mid rates, major currencies), \code{"b"} (mid rates, less common currencies),
or \code{"c"} (bid/ask rates).}

\item{code}{(\code{NULL} | \code{character(1)})\cr
ISO 4217 currency code (e.g. \code{"usd"}, \code{"eur"}). If \code{NULL}, returns all currencies.}

\item{start_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
Start date of the data.}

\item{end_date}{(\code{NULL} | \code{character(1)} | \code{Date(1)})\cr
End date of the data.}

\item{last_n}{(\code{NULL} | \code{integer(1)})\cr
Return only the last \code{n} quotations.}
}
\value{
A \code{\link[data.table:data.table]{data.table::data.table()}} with exchange rates.
}
\description{
Retrieve foreign currency exchange rates from the NBP Web API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
nbp_fx_rates("a", "eur")
}
\dontshow{\}) # examplesIf}
}
\seealso{
Other data:
\code{\link[=bbk_data]{bbk_data()}},
\code{\link[=bbk_series]{bbk_series()}},
\code{\link[=bde_data]{bde_data()}},
\code{\link[=bde_latest]{bde_latest()}},
\code{\link[=bdf_codelist]{bdf_codelist()}},
\code{\link[=bdf_data]{bdf_data()}},
\code{\link[=bdf_dataset]{bdf_dataset()}},
\code{\link[=bdp_data]{bdp_data()}},
\code{\link[=bis_data]{bis_data()}},
\code{\link[=boc_data]{boc_data()}},
\code{\link[=boe_data]{boe_data()}},
\code{\link[=boj_data]{boj_data()}},
\code{\link[=ecb_data]{ecb_data()}},
\code{\link[=nbp_gold]{nbp_gold()}},
\code{\link[=nob_data]{nob_data()}},
\code{\link[=onb_data]{onb_data()}},
\code{\link[=snb_data]{snb_data()}},
\code{\link[=srb_cross_rates]{srb_cross_rates()}},
\code{\link[=srb_data]{srb_data()}}
}
\concept{data}
